# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
from typing import (
    List,
    Tuple,
)
import numpy as np

from smarts.core.scenario import (
    Goal,
    Start,
)
from smarts.core.sensors import (
    Heading,
    Observation,
)
from ultra.baselines.common.social_vehicle_extraction import (
    get_social_vehicles_leading,
    get_social_vehicles_states_sorted_by_distance,
)
from ultra.utils.common import (
    get_closest_waypoint,
    get_path_to_goal,
)


class StatePreprocessor:
    def __call__(self, state: Observation, *args, **kwargs):
        return self._preprocess_state(state, *args, **kwargs)

    def _preprocess_state(self, state: Observation, *args, **kwargs):
        raise NotImplementedError("State preprocessing method is not defined.")

    @staticmethod
    def get_state_description(*args, **kwargs):
        raise NotImplementedError("State description is not defined.")

    @staticmethod
    def get_social_vehicles_representation(
        social_vehicles,
        social_vehicle_config,
        ego_position,
        ego_heading,
        ego_waypoints,
    ):
        if social_vehicle_config["encoder"]["use_leading_vehicles"]:
            social_vehicles = get_social_vehicles_leading(
                ego_vehicle_pos=ego_position,
                ego_vehicle_heading=ego_heading,
                neighborhood_vehicles=social_vehicles,
                waypoint_paths=ego_waypoints,
                extractor_func=social_vehicle_config["social_vehicle_extractor_func"],
                **social_vehicle_config["encoder"]["use_leading_vehicles"],
            )
        else:
            social_vehicles = get_social_vehicles_states_sorted_by_distance(
                ego_vehicle_pos=ego_position,
                ego_vehicle_heading=ego_heading,
                neighborhood_vehicles=social_vehicles,
                social_vehicle_config=social_vehicle_config,
                extractor_func=social_vehicle_config["social_vehicle_extractor_func"],
            )
        social_vehicles = np.asarray(social_vehicles).astype(np.float32)
        return social_vehicles

    @staticmethod
    def extract_ego_speed(state: Observation) -> float:
        """
        Extracts the ego vehicle's speed from a raw environment observation.
            Parameters:
                state (Observation): The raw environment observation.
            Returns:
                (float): The ego vehicle's speed in meters per second.
        """
        return state.ego_vehicle_state.speed

    @staticmethod
    def extract_ego_steering(state: Observation) -> float:
        """
        Extracts the ego vehicle's speed from a raw environment observation.
            Parameters:
                state (Observation): The raw environment observation.
            Returns:
                (float): The angle of the front wheels in radians.
        """
        return state.ego_vehicle_state.steering

    @staticmethod
    def extract_ego_position(state: Observation) -> Tuple:
        """
        Extracts the ego vehicle's position from a raw environment observation.
            Parameters:
                state (Observation): The raw environment observation.
            Returns:
                (numpy.array): A 3D NumPy array of the x, y, and z coordinates of
                    the center of the vehicle's bounding box's bottom plane.
        """
        return state.ego_vehicle_state.position

    @staticmethod
    def extract_ego_heading(state: Observation) -> Heading:
        """
        Extracts the ego vehicle's heading from a raw environment observation.
            Parameters:
                state (Observation): The raw environment observation.
            Returns:
                (Heading): A class inheriting from float, containing the vehicle's
                    heading in radians.
        """
        return state.ego_vehicle_state.heading

    @staticmethod
    def extract_ego_waypoints(state: Observation) -> List:
        """
        Extracts the ego vehicle's waypoints from a raw environment observation.
            Parameters:
                state (Observation): The raw environment observation.
            Returns:
                (List[List[Waypoint]]): A list of a list of the ego vehicle's waypoints.
        """
        return state.waypoint_paths

    @staticmethod
    def extract_ego_start(state: Observation) -> Start:
        """
        Extracts the ego vehicle's mission start from a raw environemnt observation.
            Parameters:
                state (Observation): The raw environment observation.
            Returns:
                (Start): A dataclass containing a 'position' attribute containing the
                    x and y coordinates of the start position in a tuple, and a 'heading'
                    attribute containing the Heading of the start position.
        """
        return state.ego_vehicle_state.mission.start

    @staticmethod
    def extract_ego_goal(state: Observation) -> Goal:
        """
        Extracts the ego vehicle's mission goal from a raw environment observation.
            Parameters:
                state (Observation): The raw environment observation.
            Returns:
                (PositionalGoal): A dataclass containing a 'position' attribute containing
                    the x and y coordinates of the goal in a tuple, and a float 'radius'
                    attribute containing the radius in meters the vehicle has to be in
                    order to have reached the goal
        """
        return state.ego_vehicle_state.mission.goal

    @staticmethod
    def extract_ego_goal_path(ego_goal, ego_waypoints, ego_start):
        return get_path_to_goal(goal=ego_goal, paths=ego_waypoints, start=ego_start)

    @staticmethod
    def extract_closest_waypoint(
        ego_goal_path, ego_position, ego_heading, num_lookahead=100
    ):
        return get_closest_waypoint(
            num_lookahead=num_lookahead,
            goal_path=ego_goal_path,
            ego_position=ego_position,
            ego_heading=ego_heading,
        )

    @staticmethod
    def extract_social_vehicles(state: Observation) -> List:
        """
        Extracts the social vehicles from a raw environment observation.
            Parameters:
                state (Observation): The raw environment observation.
            Returns:
                (List[VehicleObservation]): A list of vehicle observations containing
                    data about each neighboring social vehicle.
        """
        return state.neighborhood_vehicle_states



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import argparse
import glob
import json
import os
import random
import re
import sys
import timeit
from collections import defaultdict

import dill
import gym
import numpy as np
import ray
import torch
from matplotlib import pyplot as plt

from smarts.core.agent import Agent, AgentSpec
from smarts.core.agent_interface import AgentInterface, AgentType, NeighborhoodVehicles
from smarts.core.controllers import ActionSpaceType
from smarts.core.utils.episodes import episodes
from ultra.scenarios.analysis.base_analysis import BaseAnalysis
from ultra.scenarios.common.social_vehicle_definitions import get_social_vehicle_color
from ultra.scenarios.common.visualization import (
    convert_to_gif,
    draw_intersection,
    profile_vehicles,
)
from ultra.scenarios.generate_scenarios import get_direction

num_gpus = 1 if torch.cuda.is_available() else 0


@ray.remote(num_gpus=num_gpus / 2)
class BehaviorAnalysis(BaseAnalysis):
    def __init__(self):
        super(BaseAnalysis, self).__init__()
        self.analysis = []

    def draw_plots(self, save_dir, failure_time=None):
        start = 0
        end = 600
        behaviors = {}
        for scenario in self.analysis:
            for behavior in scenario.keys():
                if behavior not in behaviors:
                    behaviors[behavior] = {
                        "avg_speed_junction": [],
                        "min_speed_junction": [],
                        "max_speed_junction": [],
                        "avg_accel_junction": [],
                        "min_accel_junction": [],
                        "max_accel_junction": [],
                        "total_steps_junction": [],
                        "exceeds_speed_limit": 0,
                    }
                s, e = scenario[behavior]["in_junction"][0]
                behaviors[behavior]["avg_speed_junction"].append(
                    np.mean(scenario[behavior]["speeds"][s:e])
                )
                behaviors[behavior]["min_speed_junction"].append(
                    scenario[behavior]["min_speed_junction"][0]
                )
                behaviors[behavior]["max_speed_junction"].append(
                    scenario[behavior]["max_speed_junction"][0]
                )
                behaviors[behavior]["avg_accel_junction"].append(
                    np.mean(scenario[behavior]["accels"][s:e])
                )
                behaviors[behavior]["min_accel_junction"].append(
                    scenario[behavior]["min_accel_junction"][0]
                )
                behaviors[behavior]["max_accel_junction"].append(
                    scenario[behavior]["max_accel_junction"][0]
                )
                behaviors[behavior]["total_steps_junction"].append(
                    scenario[behavior]["steps_in_junction"][0]
                )

        for key in [
            "min_speed_junction",
            "max_speed_junction",
            "avg_speed_junction",
            "min_accel_junction",
            "max_accel_junction",
            "avg_accel_junction",
            "total_steps_junction",
        ]:
            plt.figure()
            if "speed" in key:
                bins = [i for i in range(40)]
            elif "accel" in key:
                bins = [i for i in range(-15, 15)]
            else:
                bins = [i for i in range(1000)]
            for behavior, data in behaviors.items():
                n, bins, patches = plt.hist(
                    x=data[key], bins=bins, color="#0504aa", alpha=0.7, rwidth=0.85
                )
                plt.grid(axis="y", alpha=0.75)
                plt.xlabel(key)
                plt.title(behavior)
                plt.ylabel("Frequency")
                maxfreq = n.max()
                # Set a clean upper y-axis limit.
                plt.ylim(
                    ymax=np.ceil(maxfreq / 10) * 10 if maxfreq % 10 else maxfreq + 10
                )
                plt.savefig(f"{save_dir}/{behavior}_{key}.png")
                plt.close()

    def analyze(self, episode_time, stopwatcher_max_steps, simulation_data):
        analysis = {}
        for v_id, v_state in self.social_vehicles_states.items():
            if "stopwatcher" in v_state["behavior"]:
                _, behavior = v_state["behavior"].split("_")
                if behavior not in self.analysis:
                    analysis[behavior] = {
                        "max_steps": [],
                        "steps_in_junction": [],
                        "max_road_speed": None,
                        "speeds": [],
                        "accels": [],
                        "min_speed_junction": [],
                        "max_speed_junction": [],
                        "min_accel_junction": [],
                        "max_accel_junction": [],
                        "in_junction": [],
                        "steps_in_junction": [],
                    }
                in_junction = v_state["in_junction"]
                speed_in_junction = v_state["speeds"][in_junction[0] : in_junction[1]]
                accel_in_junction = v_state["accels"][in_junction[0] : in_junction[1]]
                analysis[behavior]["in_junction"].append(in_junction)
                analysis[behavior]["min_speed_junction"].append(min(speed_in_junction))
                analysis[behavior]["max_speed_junction"].append(max(speed_in_junction))
                analysis[behavior]["min_accel_junction"].append(min(accel_in_junction))
                analysis[behavior]["max_accel_junction"].append(max(accel_in_junction))
                print("in junction", in_junction)
                analysis[behavior]["max_steps"].append(v_state["steps"])
                analysis[behavior]["steps_in_junction"].append(
                    int(in_junction[1]) - int(in_junction[0])
                )
                analysis[behavior][
                    "max_road_speed"
                ] = 19.44  # road_speads[v_state["route"]]
                analysis[behavior]["speeds"] = v_state["speeds"]
                analysis[behavior]["accels"] = v_state["accels"]
        self.analysis.append(analysis)

    def save_data(self, save_dir):
        with open(f"{save_dir}/analysis.pkl", "wb") as handle:
            dill.dump(self.analysis, handle)
        print(f"saved data to {save_dir}/analysis.pkl")

    def load_data(self, path):
        with open(path, "rb") as handle:
            self.analysis = dill.load(handle)
        print(f"loaded data {len(self.analysis)} from {path}")

    def run(self, **params):
        super().run(**params, analyze_func=self.analyze)



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
from itertools import cycle
import glob, yaml
from smarts.core.scenario import Scenario
from smarts.env.rllib_hiway_env import RLlibHiWayEnv
from ultra.baselines.adapter import BaselineAdapter
from ultra.baselines.common.yaml_loader import load_yaml
import numpy as np
from scipy.spatial import distance
import math, os
from sys import path

path.append("./ultra")
from ultra.utils.common import (
    get_closest_waypoint,
    get_path_to_goal,
    ego_social_safety,
)


class RLlibUltraEnv(RLlibHiWayEnv):
    def __init__(self, config):
        print(">>> ENV: ", config["eval_mode"])
        self.scenario_info = config["scenario_info"]
        self.scenarios = self.get_task(self.scenario_info[0], self.scenario_info[1])
        self._timestep_sec = config["timestep_sec"]
        self._seed = config["seed"]
        if not config["eval_mode"]:
            _scenarios = glob.glob(f"{self.scenarios['train']}")
        else:
            _scenarios = glob.glob(f"{self.scenarios['test']}")

        config["scenarios"] = _scenarios
        self.ultra_scores = BaselineAdapter.reward_adapter
        super().__init__(config=config)

        if config["ordered_scenarios"]:
            scenario_roots = []
            for root in config["scenarios"]:
                if Scenario.is_valid_scenario(root):
                    # The case that this is a scenario root
                    scenario_roots.append(root)
                else:
                    # The case that there this is a directory of scenarios: find each of the roots
                    scenario_roots.extend(Scenario.discover_scenarios(root))
            # Also see `smarts.env.HiwayEnv`
            self._scenarios_iterator = cycle(
                Scenario.variations_for_all_scenario_roots(
                    scenario_roots, list(config["agent_specs"].keys())
                )
            )

    def generate_logs(self, observation, highwayenv_score):
        ego_state = observation.ego_vehicle_state
        start = observation.ego_vehicle_state.mission.start
        goal = observation.ego_vehicle_state.mission.goal
        path = get_path_to_goal(
            goal=goal, paths=observation.waypoint_paths, start=start
        )
        closest_wp, _ = get_closest_waypoint(
            num_lookahead=100,
            goal_path=path,
            ego_position=ego_state.position,
            ego_heading=ego_state.heading,
        )
        signed_dist_from_center = closest_wp.signed_lateral_error(ego_state.position)
        lane_width = closest_wp.lane_width * 0.5
        ego_dist_center = signed_dist_from_center / lane_width

        linear_jerk = np.linalg.norm(ego_state.linear_jerk)
        angular_jerk = np.linalg.norm(ego_state.angular_jerk)

        # Distance to goal
        ego_2d_position = ego_state.position[0:2]
        goal_dist = distance.euclidean(ego_2d_position, goal.position)

        angle_error = closest_wp.relative_heading(
            ego_state.heading
        )  # relative heading radians [-pi, pi]

        # number of violations
        (ego_num_violations, social_num_violations,) = ego_social_safety(
            observation,
            d_min_ego=1.0,
            t_c_ego=1.0,
            d_min_social=1.0,
            t_c_social=1.0,
            ignore_vehicle_behind=True,
        )

        info = dict(
            position=ego_state.position,
            speed=ego_state.speed,
            steering=ego_state.steering,
            heading=ego_state.heading,
            dist_center=abs(ego_dist_center),
            start=start,
            goal=goal,
            closest_wp=closest_wp,
            events=observation.events,
            ego_num_violations=ego_num_violations,
            social_num_violations=social_num_violations,
            goal_dist=goal_dist,
            linear_jerk=np.linalg.norm(ego_state.linear_jerk),
            angular_jerk=np.linalg.norm(ego_state.angular_jerk),
            env_score=self.ultra_scores(observation, highwayenv_score),
        )
        return info

    def step(self, agent_actions):
        agent_actions = {
            agent_id: self._agent_specs[agent_id].action_adapter(action)
            for agent_id, action in agent_actions.items()
        }

        observations, rewards, agent_dones, extras = self._smarts.step(agent_actions)

        observations = {
            agent_id: obs
            for agent_id, obs in observations.items()
            if agent_id in agent_actions
        }
        rewards = {
            agent_id: reward
            for agent_id, reward in rewards.items()
            if agent_id in agent_actions
        }
        scores = {
            agent_id: score
            for agent_id, score in extras["scores"].items()
            if agent_id in agent_actions
        }

        infos = {
            agent_id: {"score": value, "env_obs": observations[agent_id]}
            for agent_id, value in extras["scores"].items()
        }

        # Ensure all contain the same agent_ids as keys
        assert (
            agent_actions.keys()
            == observations.keys()
            == rewards.keys()
            == infos.keys()
        )

        for agent_id in observations:
            agent_spec = self._agent_specs[agent_id]
            observation = observations[agent_id]
            reward = rewards[agent_id]
            info = infos[agent_id]

            rewards[agent_id] = agent_spec.reward_adapter(observation, reward)
            observations[agent_id] = agent_spec.observation_adapter(observation)
            infos[agent_id] = agent_spec.info_adapter(observation, reward, info)
            infos[agent_id]["logs"] = self.generate_logs(observation, reward)

        for done in agent_dones.values():
            self._dones_registered += 1 if done else 0

        agent_dones["__all__"] = self._dones_registered == len(self._agent_specs)

        return observations, rewards, agent_dones, infos

    def get_task(self, task_id, task_level):
        base_dir = os.path.join(os.path.dirname(__file__), "../")
        config_path = os.path.join(base_dir, "config.yaml")

        with open(config_path, "r") as task_file:
            scenarios = yaml.safe_load(task_file)["tasks"]
            task = scenarios[f"task{task_id}"][task_level]

        task["train"] = os.path.join(base_dir, task["train"])
        task["test"] = os.path.join(base_dir, task["test"])
        return task

    @property
    def info(self):
        return {
            "scenario_info": self.scenario_info,
            "timestep_sec": self._timestep_sec,
            "headless": self._headless,
        }

    @property
    def seed(self):
        return self._seed



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# Inspired by https://github.com/fxia22/pointnet.pytorch/blob/master/pointnet/model.py
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
import numpy as np


class STNkd(nn.Module):
    def __init__(self, k=64, nc=16):
        super(STNkd, self).__init__()
        self.conv1 = torch.nn.Conv1d(k, nc, 1)
        self.conv2 = torch.nn.Conv1d(nc, nc * 4, 1)
        self.conv3 = torch.nn.Conv1d(nc * 4, nc * 16, 1)
        self.fc1 = nn.Linear(nc * 16, nc * 4)
        self.fc2 = nn.Linear(nc * 4, nc)
        self.fc3 = nn.Linear(nc, k * k)
        self.relu = nn.ReLU()

        # self.bn1 = nn.BatchNorm1d(nc)
        # self.bn2 = nn.BatchNorm1d(nc * 4)
        # self.bn3 = nn.BatchNorm1d(nc * 16)
        # self.bn4 = nn.BatchNorm1d(nc * 4)
        # self.bn5 = nn.BatchNorm1d(nc)
        identity = lambda x: x
        self.bn1, self.bn2, self.bn3, self.bn4, self.bn5 = [identity] * 5

        self.k = k
        self.nc = nc

    def forward(self, x):
        batchsize = x.size()[0]
        x = F.relu(self.bn1(self.conv1(x)))
        x = F.relu(self.bn2(self.conv2(x)))
        x = F.relu(self.bn3(self.conv3(x)))
        x = torch.max(x, 2, keepdim=True)[0]
        x = x.view(-1, self.nc * 16)

        x = F.relu(self.bn4(self.fc1(x)))
        x = F.relu(self.bn5(self.fc2(x)))
        x = self.fc3(x)

        iden = (
            Variable(torch.from_numpy(np.eye(self.k).flatten().astype(np.float32)))
            .view(1, self.k * self.k)
            .repeat(batchsize, 1)
        )
        if x.is_cuda:
            iden = iden.cuda()
        x = x + iden
        x = x.view(-1, self.k, self.k)
        return x


class PNEncoderBatched(nn.Module):
    def __init__(
        self,
        input_dim=3,
        global_features=True,
        feature_transform=True,
        nc=16,
        transform_loss_weight=0.1,
    ):
        assert global_features
        super(PNEncoderBatched, self).__init__()
        self.input_dim = input_dim
        self.nc = nc
        self.global_features = global_features
        self.feature_transform = feature_transform
        self.transform_loss_weight = transform_loss_weight

        self.transformD = STNkd(k=input_dim, nc=nc)
        self.conv1 = nn.Conv1d(self.input_dim, nc, 1)
        # self.bn1 = nn.BatchNorm1d(nc)
        self.conv2 = nn.Conv1d(nc, nc * 4, 1)
        # self.bn2 = nn.BatchNorm1d(nc * 4)
        self.conv3 = nn.Conv1d(nc * 4, nc * 16, 1)
        # self.bn3 = nn.BatchNorm1d(nc * 16)
        identity = lambda x: x
        self.bn1, self.bn2, self.bn3 = [identity] * 3
        # self.bn1, self.bn2, self.bn3 = nn.LayerNorm(nc), nn.LayerNorm(nc * 4), nn.LayerNorm(nc * 16)

        self.output_dim = nc * 16

        if self.feature_transform:
            self.transformF = STNkd(k=self.nc, nc=nc)

        self.empty_fill = torch.from_numpy(
            np.asarray(
                [
                    -1,
                    -1,  # bottom left very far from ego
                    -1,
                    0,  # move backward from ego
                ]
            )
        ).unsqueeze(0)

    def transform_loss(self, transD, transF):
        transform_loss_raw_number = [
            [
                feature_transform_regularizer(e.cpu()).to(e.device)
                if e is not None
                else 0.0
                for e in transD
            ],
            [
                feature_transform_regularizer(e.cpu()).to(e.device)
                if e is not None
                else 0.0
                for e in transF
            ],
        ]
        transform_loss = [
            sum(transform_loss_raw_number[0]) / len(transform_loss_raw_number[0]),
            sum(transform_loss_raw_number[1]) / len(transform_loss_raw_number[1]),
        ]
        mean_transform_loss = sum(transform_loss) / len(transform_loss)
        aux_losses = {
            "transform": {
                "value": mean_transform_loss,
                "weight": self.transform_loss_weight,
            }
        }
        return aux_losses

    def forward(self, social_vehicles_state, training=False):
        max_num_social_in_batch = max([len(e) for e in social_vehicles_state])
        max_num_social_in_batch = max(max_num_social_in_batch, 1)
        # because PointNet is taking the trajectory-wise max
        # padding with a bunch of zero here wouldn't change the final training result
        # if training is proper (need to test)
        # also, it is possible to pad with any amount of zeros
        self.empty_fill = self.empty_fill.to(social_vehicles_state[0].device)
        input_tensor = torch.zeros(
            len(social_vehicles_state),
            max_num_social_in_batch,
            social_vehicles_state[0].shape[-1],
            device=social_vehicles_state[0].device,
        )
        for j in range(len(social_vehicles_state)):
            input_tensor[j, : len(social_vehicles_state[j])] = social_vehicles_state[j]
            # if want to fill with zero, comment line below or set self.empty_fil to 0
            input_tensor[j, len(social_vehicles_state[j]) :] = self.empty_fill
        social_features, transD, transF = self._forward(input_tensor)
        social_features = [e.unsqueeze(0) for e in social_features]
        transD = [e.unsqueeze(0) for e in transD]
        transF = [e.unsqueeze(0) for e in transF]
        if training:
            aux_losses = self.transform_loss(transD, transF)
            return social_features, aux_losses
        else:
            return social_features, {}

    def _forward(self, points):
        points = points.transpose(-1, -2)
        transD = self.transformD(points)
        x = points.transpose(2, 1)
        x = torch.bmm(x, transD)
        x = x.transpose(2, 1)
        x = F.relu(self.bn1(self.conv1(x)))

        if self.feature_transform:
            transF = self.transformF(x)
            x = x.transpose(2, 1)
            x = torch.bmm(x, transF)
            x = x.transpose(2, 1)
        else:
            transF = None

        point_features = x
        x = F.relu(self.bn2(self.conv2(x)))
        x = self.bn3(self.conv3(x))
        x = torch.max(x, 2, keepdim=True)[0]
        x = x.view(-1, self.output_dim)

        if self.global_features:
            return x, transD, transF
        else:
            N = points.size()[2]
            x = x.view(-1, self.nc, 1).repeat(1, 1, N)
            return torch.cat([x, point_features], 1), transD, transF


def feature_transform_regularizer(trans):
    d = trans.size()[1]
    batchsize = trans.size()[0]
    I = torch.eye(d)[None, :, :]
    if trans.is_cuda:
        I = I.cuda()
    loss = torch.mean(
        torch.norm(torch.bmm(trans, trans.transpose(2, 1)) - I, dim=(1, 2))
    )
    if torch.isnan(loss):
        loss = 0.0
    return loss



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import torch, gym
from torch import nn
from torch.distributions.normal import Normal
from ray.rllib.models.torch.torch_modelv2 import TorchModelV2
from ray.rllib.models.torch.fcnet import FullyConnectedNetwork as TorchFCNet
from ultra.baselines.common.state_preprocessor import *


class CustomFCModel(TorchModelV2, nn.Module):
    """Example of interpreting repeated observations."""

    def __init__(
        self,
        obs_space: gym.spaces.Space,
        action_space: gym.spaces.Space,
        num_outputs: int,
        model_config,
        name: str,
        **customized_model_kwargs
    ):
        super(CustomFCModel, self).__init__(
            obs_space=obs_space,
            action_space=action_space,
            num_outputs=num_outputs,
            model_config=model_config,
            name=name,
        )
        nn.Module.__init__(self)

        if "adapter" in model_config["custom_model_config"]:
            adapter = model_config["custom_model_config"]["adapter"]
        else:
            adapter = customized_model_kwargs["adapter"]

        social_feature_encoder_class = adapter.social_feature_encoder_class
        social_feature_encoder_params = adapter.social_feature_encoder_params
        self.social_feature_encoder = (
            social_feature_encoder_class(**social_feature_encoder_params)
            if social_feature_encoder_class
            else None
        )

        self.model = TorchFCNet(
            obs_space, action_space, num_outputs, model_config, name
        )

    def forward(self, input_dict, state, seq_lens):

        low_dim_state = input_dict["obs"]["low_dim_states"]
        social_vehicles_state = input_dict["obs"]["social_vehicles"]

        social_feature = []
        if self.social_feature_encoder is not None:
            social_feature, _ = self.social_feature_encoder(social_vehicles_state)
        else:
            social_feature = [e.reshape(1, -1) for e in social_vehicles_state]

        input_dict["obs"]["social_vehicles"] = (
            torch.cat(social_feature, 0) if len(social_feature) > 0 else []
        )
        return self.model.forward(input_dict, state, seq_lens)

    def value_function(self):
        return self.model.value_function()



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# some parts of this implementation is inspired by https://github.com/Khrylx/PyTorch-RL
import torch, os, yaml
import numpy as np
from ultra.baselines.ppo.ppo.network import PPONetwork
from smarts.core.agent import Agent
import pathlib
import os
from ultra.utils.common import (
    compute_sum_aux_losses,
    normalize_im,
    to_3d_action,
    to_2d_action,
)
from ultra.baselines.common.yaml_loader import load_yaml
from ultra.baselines.common.social_vehicle_config import get_social_vehicle_configs
from ultra.baselines.common.baseline_state_preprocessor import BaselineStatePreprocessor


class PPOPolicy(Agent):
    def __init__(
        self,
        policy_params=None,
        checkpoint_dir=None,
    ):
        self.policy_params = policy_params
        self.batch_size = int(policy_params["batch_size"])
        self.hidden_units = int(policy_params["hidden_units"])
        self.mini_batch_size = int(policy_params["mini_batch_size"])
        self.epoch_count = int(policy_params["epoch_count"])
        self.gamma = float(policy_params["gamma"])
        self.l = float(policy_params["l"])
        self.eps = float(policy_params["eps"])
        self.actor_tau = float(policy_params["actor_tau"])
        self.critic_tau = float(policy_params["critic_tau"])
        self.entropy_tau = float(policy_params["entropy_tau"])
        self.logging_freq = int(policy_params["logging_freq"])
        self.current_iteration = 0
        self.current_log_prob = None
        self.current_value = None
        self.seed = int(policy_params["seed"])
        self.lr = float(policy_params["lr"])
        self.log_probs = []
        self.values = []
        self.rewards = []
        self.actions = []
        self.states = []
        self.terminals = []
        self.action_size = int(policy_params["action_size"])
        self.prev_action = np.zeros(self.action_size)

        # state preprocessing
        self.social_policy_hidden_units = int(
            policy_params["social_vehicles"].get("social_policy_hidden_units", 0)
        )
        self.social_capacity = int(
            policy_params["social_vehicles"].get("social_capacity", 0)
        )
        self.observation_num_lookahead = int(
            policy_params.get("observation_num_lookahead", 0)
        )
        self.social_policy_init_std = int(
            policy_params["social_vehicles"].get("social_policy_init_std", 0)
        )
        self.num_social_features = int(
            policy_params["social_vehicles"].get("num_social_features", 0)
        )
        self.social_vehicle_config = get_social_vehicle_configs(
            **policy_params["social_vehicles"]
        )

        self.social_vehicle_encoder = self.social_vehicle_config["encoder"]
        self.state_description = BaselineStatePreprocessor.get_state_description(
            policy_params["social_vehicles"],
            policy_params["observation_num_lookahead"],
            self.action_size,
        )
        # self.state_preprocessor = StatePreprocessor(
        #     preprocess_state, to_2d_action, self.state_description
        # )
        self.social_feature_encoder_class = self.social_vehicle_encoder[
            "social_feature_encoder_class"
        ]
        self.social_feature_encoder_params = self.social_vehicle_encoder[
            "social_feature_encoder_params"
        ]

        # others
        self.checkpoint_dir = checkpoint_dir
        self.device_name = "cuda:0" if torch.cuda.is_available() else "cpu"
        self.device = torch.device(self.device_name)
        self.save_codes = (
            policy_params["save_codes"] if "save_codes" in policy_params else None
        )

        # PPO
        self.ppo_net = PPONetwork(
            self.action_size,
            self.state_size,
            hidden_units=self.hidden_units,
            init_std=self.social_policy_init_std,
            seed=self.seed,
            social_feature_encoder_class=self.social_feature_encoder_class,
            social_feature_encoder_params=self.social_feature_encoder_params,
        ).to(self.device)
        self.optimizer = torch.optim.Adam(self.ppo_net.parameters(), lr=self.lr)
        self.step_count = 0
        if self.checkpoint_dir:
            self.load(self.checkpoint_dir)

    @property
    def state_size(self):
        # Adjusting state_size based on number of features (ego+social)
        size = sum(self.state_description["low_dim_states"].values())
        if self.social_feature_encoder_class:
            size += self.social_feature_encoder_class(
                **self.social_feature_encoder_params
            ).output_dim
        else:
            size += self.social_capacity * self.num_social_features

        # adding the previous action
        size += self.action_size
        return size

    def act(self, state, explore=True):
        state["low_dim_states"] = np.float32(
            np.append(state["low_dim_states"], self.prev_action)
        )
        state["social_vehicles"] = (
            torch.from_numpy(state["social_vehicles"]).unsqueeze(0).to(self.device)
        )
        state["low_dim_states"] = (
            torch.from_numpy(state["low_dim_states"]).unsqueeze(0).to(self.device)
        )

        with torch.no_grad():
            dist, value = self.ppo_net(x=state)
        if explore:  # training mode
            action = dist.sample()
            log_prob = dist.log_prob(action)

            self.current_log_prob = log_prob
            self.current_value = value

            action = torch.squeeze(action)
            action = action.data.cpu().numpy()
        else:  # testing mode
            mean = torch.squeeze(dist.loc)
            action = mean.data.cpu().numpy()
        self.step_count += 1
        return to_3d_action(action)

    def step(self, state, action, reward, next_state, done, info):
        # dont treat timeout as done equal to True
        max_steps_reached = info["logs"]["events"].reached_max_episode_steps
        if max_steps_reached:
            done = False
        action = to_2d_action(action)

        # pass social_vehicle_rep through the network
        # state['low_dim_states'] = torch.from_numpy(np.float32(np.append(state['low_dim_states'],self.prev_action))).unsqueeze(0)

        self.log_probs.append(self.current_log_prob.to(self.device))
        self.values.append(self.current_value.to(self.device))
        self.states.append(state)
        self.rewards.append(torch.FloatTensor([reward]).to(self.device))
        self.actions.append(
            torch.FloatTensor(
                action.reshape(
                    self.action_size,
                )
            ).to(self.device)
        )
        self.terminals.append(1.0 - float(done * 1))

        output = {}
        # batch updates over multiple episodes
        if len(self.terminals) >= self.batch_size:
            output = self.learn()

        self.prev_action = action if not done else np.zeros(self.action_size)
        return output

    def compute_returns(self, rewards, masks, values, gamma=0.99, lamda=0.95):
        """This computes the lambda return using values, rewards, and indication for terminal states.
        Source of iterative form (The Generalized Advantage Estimator):
        https://danieltakeshi.github.io/2017/04/02/notes-on-the-generalized-advantage-estimation-paper/
        """
        values = values + [0]
        A_GAE = 0
        returns = []
        for step in reversed(range(len(rewards))):
            delta = (
                rewards[step] - values[step] + gamma * values[step + 1] * masks[step]
            )
            A_GAE = (
                delta + gamma * lamda * A_GAE * masks[step]
            )  #  (γλ)^l * delta_{t+1} = Advantage
            # but we need to return the returns G_t{λ}, so we add to that V_t before returning
            returns.insert(0, A_GAE + values[step])
        return returns

    def make_state_from_dict(self, states, device):
        low_dim_states = (
            torch.cat([e["low_dim_states"] for e in states], dim=0).float().to(device)
        )
        social_vehicles = [e["social_vehicles"] for e in states]

        out = {
            "low_dim_states": low_dim_states,
            "social_vehicles": social_vehicles,
        }

        return out

    def get_minibatch(
        self, mini_batch_size, states, actions, log_probs, returns, advantage
    ):
        """Generator that can give the next batch in the dataset.
        This returns a minibatch of bunch of tensors
        """
        batch_size = len(states)
        ids = np.random.permutation(batch_size)
        whole_mini_batchs = batch_size // mini_batch_size * mini_batch_size
        no_mini_batchs = batch_size // mini_batch_size

        # split the dataset into number of minibatchs and discard the rest
        # (ex. if you have mini_batch=32 and batch_size = 100 then the utilized portion is only 96=3*32)
        splits = np.split(ids[:whole_mini_batchs], no_mini_batchs)

        # using a generator to return different mini-batch each time.
        for i in range(len(splits)):
            states_mini_batch = [states[e] for e in splits[i]]
            states_mini_batch = self.make_state_from_dict(
                states_mini_batch, device=self.device
            )

            yield (
                states_mini_batch,
                actions[splits[i], :].to(self.device),
                log_probs[splits[i], :].to(self.device),
                returns[splits[i], :].to(self.device),
                advantage[splits[i], :].to(self.device),
            )

    def get_ratio(self, new_pi_log_probs, old_pi_log_probs):
        return (new_pi_log_probs - old_pi_log_probs).exp()

    def update(
        self, n_epochs, mini_batch_size, states, actions, log_probs, returns, advantages
    ):
        total_actor_loss = 0
        total_critic_loss = 0
        total_entropy_loss = 0
        # multiple epochs
        for _ in range(n_epochs):
            # minibatch updates
            for (
                state,
                action,
                old_pi_log_probs,
                return_batch,
                advantage,
            ) in self.get_minibatch(
                mini_batch_size, states, actions, log_probs, returns, advantages
            ):
                (dist, value), aux_losses = self.ppo_net(state, training=True)
                entropy = dist.entropy().mean()  # L_S
                new_pi_log_probs = dist.log_prob(action)

                ratio = self.get_ratio(new_pi_log_probs, old_pi_log_probs)
                L_CPI = ratio * advantage
                clipped_version = (
                    torch.clamp(ratio, 1.0 - self.eps, 1.0 + self.eps) * advantage
                )

                # loss and clipping
                actor_loss = -torch.min(L_CPI, clipped_version).mean()  # L_CLIP
                critic_loss = (
                    (return_batch - value).pow(2).mean()
                )  # L_VF (squared error loss)

                aux_losses = compute_sum_aux_losses(aux_losses)

                # overall loss
                loss = (
                    self.critic_tau * critic_loss
                    + self.actor_tau * actor_loss
                    - self.entropy_tau * entropy
                    + aux_losses
                )

                # calculate gradients and update the weights
                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()

                total_actor_loss += actor_loss.item()
                total_critic_loss += critic_loss.item()
                total_entropy_loss += entropy.item()

        average_actor_loss = total_actor_loss / (
            n_epochs * (self.batch_size / self.mini_batch_size)
        )
        average_critic_loss = total_critic_loss / (
            n_epochs * (self.batch_size / self.mini_batch_size)
        )
        average_entropy_loss = total_entropy_loss / (
            n_epochs * (self.batch_size / self.mini_batch_size)
        )

        output = {
            "loss/critic": {
                "type": "scalar",
                "data": average_critic_loss,
                "freq": self.logging_freq,
            },
            "loss/actor": {
                "type": "scalar",
                "data": average_actor_loss,
                "freq": self.logging_freq,
            },
            "loss/entropy": {
                "type": "scalar",
                "data": average_entropy_loss,
                "freq": self.logging_freq,
            },
        }

        return output

    def learn(self):

        # compute lambda returns from (rewards, values) trajectories:
        returns = self.compute_returns(
            self.rewards, self.terminals, self.values, self.gamma, self.l
        )

        # convert lists into Pytorch tensors
        states = self.states
        actions = torch.stack(self.actions)
        log_probs = torch.cat(self.log_probs).detach()
        values = torch.cat(self.values).detach()
        returns = torch.cat(returns).detach()
        advantages = returns - values

        # normalize advatages
        advantages = (advantages - advantages.mean()) / (advantages.std() + 1e-5)

        # PPO update for # of epochs and over # of minibatchs

        output = self.update(
            self.epoch_count,
            self.mini_batch_size,
            states,
            actions,
            log_probs,
            returns,
            advantages,
        )

        # remove previous experiences, in preparing for the next iteration
        self.log_probs = []
        self.values = []
        self.rewards = []
        self.actions = []
        self.states = []
        self.terminals = []

        # increase the current number of iterations
        self.current_iteration += 1
        return output

    def save(self, model_dir):
        model_dir = pathlib.Path(model_dir)
        torch.save(self.ppo_net.state_dict(), model_dir / "ppo_network.pth")

    def load(self, model_dir):
        print("model loaded:", model_dir)
        model_dir = pathlib.Path(model_dir)
        map_location = None
        if self.device and self.device.type == "cpu":
            map_location = "cpu"
        self.ppo_net.load_state_dict(
            torch.load(model_dir / "ppo_network.pth", map_location=map_location)
        )

    def reset(self):
        pass



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import os, gym
from ultra.utils.ray import default_ray_kwargs
import timeit, datetime

# Set environment to better support Ray
os.environ["MKL_NUM_THREADS"] = "1"
import time
import psutil, dill, torch, inspect
import ray, torch, argparse
import numpy as np
from ray import tune
from smarts.zoo.registry import make
from ultra.env.rllib_ultra_env import RLlibUltraEnv

from ray.rllib.models import ModelCatalog

from smarts.core.controllers import ActionSpaceType
from smarts.core.agent_interface import (
    AgentInterface,
    OGM,
    Waypoints,
    NeighborhoodVehicles,
)

from ultra.baselines.rllib.models.fc_network import CustomFCModel
from ultra.baselines.rllib.agent import RllibAgent
from ultra.baselines.common.yaml_loader import load_yaml
from smarts.core.agent import AgentSpec
from ultra.baselines.adapter import BaselineAdapter

from ultra.utils.episode import Callbacks
from ultra.utils.episode import log_creator

num_gpus = 1 if torch.cuda.is_available() else 0


def train(
    task,
    num_episodes,
    policy,
    eval_info,
    timestep_sec,
    headless,
    seed,
    training_batch_size,
    log_dir,
):

    agent_name = policy
    adapter = BaselineAdapter(agent_name)
    ModelCatalog.register_custom_model("fc_model", CustomFCModel)
    config = RllibAgent.rllib_default_config(agent_name)

    rllib_policies = {
        "default_policy": (
            None,
            adapter.observation_space,
            adapter.action_space,
            {
                "model": {
                    "custom_model": "fc_model",
                    "custom_model_config": {"adapter": adapter},
                }
            },
        )
    }
    agent_specs = {
        "AGENT-007": AgentSpec(
            interface=AgentInterface(
                waypoints=Waypoints(lookahead=20),
                neighborhood_vehicles=NeighborhoodVehicles(200),
                action=ActionSpaceType.Continuous,
                rgb=False,
                max_episode_steps=600,
                debug=True,
            ),
            agent_params={},
            agent_builder=None,
            observation_adapter=adapter.observation_adapter,
            reward_adapter=adapter.reward_adapter,
            # action_adapter=adapter.action_adapter,
        )
    }

    tune_config = {
        "env": RLlibUltraEnv,
        "log_level": "WARN",
        "callbacks": Callbacks,
        "framework": "torch",
        "num_workers": 1,
        "train_batch_size": training_batch_size,  # Debugging value
        "in_evaluation": True,
        "evaluation_num_episodes": eval_info["eval_episodes"],
        "evaluation_interval": eval_info[
            "eval_rate"
        ],  # Evaluation occurs after # of eval-intervals (episodes)
        "evaluation_config": {
            "env_config": {
                "seed": seed,
                "scenario_info": task,
                "headless": headless,
                "eval_mode": True,
                "ordered_scenarios": False,
                "agent_specs": agent_specs,
                "timestep_sec": timestep_sec,
            },
            "explore": False,
        },
        "env_config": {
            "seed": seed,
            "scenario_info": task,
            "headless": headless,
            "eval_mode": False,
            "ordered_scenarios": False,
            "agent_specs": agent_specs,
            "timestep_sec": timestep_sec,
        },
        "multiagent": {"policies": rllib_policies},
    }

    config.update(tune_config)
    agent = RllibAgent(
        agent_name=agent_name,
        env=RLlibUltraEnv,
        config=tune_config,
        logger_creator=log_creator(log_dir),
    )

    # Iteration value in trainer.py (self._iterations) is the technically the number of episodes
    for i in range(num_episodes):
        results = agent.train()
        agent.log_evaluation_metrics(
            results
        )  # Evaluation metrics will now be displayed on Tensorboard


if __name__ == "__main__":
    parser = argparse.ArgumentParser("intersection-single-agent")
    parser.add_argument(
        "--task", help="Tasks available : [0, 1, 2, 3]", type=str, default="1"
    )
    parser.add_argument(
        "--level",
        help="Tasks available : [easy, medium, hard, no-traffic]",
        type=str,
        default="easy",
    )
    parser.add_argument(
        "--policy",
        help="Policies avaliable : [ppo, ddpg, td3]",
        type=str,
        default="ppo",
    )
    parser.add_argument(
        "--episodes", help="number of training episodes", type=int, default=1000000
    )
    parser.add_argument(
        "--timestep", help="environment timestep (sec)", type=float, default=0.1
    )
    parser.add_argument(
        "--headless", help="run without envision", type=bool, default=False
    )
    parser.add_argument(
        "--eval-episodes", help="number of evaluation episodes", type=int, default=100
    )
    parser.add_argument(
        "--eval-rate",
        help="run evaluation every 'n' number of iterations",
        type=int,
        default=100,
    )
    parser.add_argument(
        "--seed",
        help="environment seed",
        default=2,
        type=int,
    )
    parser.add_argument(
        "--training-batch-size",
        help="maximum samples for each training iteration",
        default=4000,
        type=int,
    )
    parser.add_argument(
        "--log-dir",
        help="Log directory location",
        default="logs",
        type=str,
    )

    args = parser.parse_args()

    ray.init()
    train(
        task=(args.task, args.level),
        num_episodes=int(args.episodes),
        policy=args.policy,
        eval_info={
            "eval_rate": int(args.eval_rate),
            "eval_episodes": int(args.eval_episodes),
        },
        timestep_sec=float(args.timestep),
        headless=args.headless,
        seed=args.seed,
        training_batch_size=args.training_batch_size,
        log_dir=args.log_dir,
    )


# keeping for later
# from ray.tune.schedulers import PopulationBasedTraining
# pbt = PopulationBasedTraining(
#     time_attr="time_total_s",
#     metric="episode_reward_mean",
#     mode="max",
#     perturbation_interval=300,
#     resample_probability=0.25,
#     hyperparam_mutations={
#         "lr": [1e-3, 5e-4, 1e-4, 5e-5, 1e-5],
#         "rollout_fragment_length": lambda: 200,
#         "train_batch_size": lambda: 2000,
#     },
# )

# analysis = tune.run(
#     "PPO",  # "SAC"
#     name="exp_1",
#     stop={"training_iteration": 10},  # {"timesteps_total": 1200},
#     checkpoint_freq=10,
#     checkpoint_at_end=True,
#     local_dir=str(result_dir),
#     resume=False,
#     restore=None,
#     max_failures=1,
#     num_samples=1,
#     export_formats=["model", "checkpoint"],
#     config=config,
#     loggers=
#     # scheduler=pbt,
#     # "lr": tune.grid_search([1e-3,1e-4])
# )



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import unittest, ray, os, sys
from ultra.rllib_train import train
import shutil

AGENT_ID = "001"
seed = 2


class RLlibTrainTest(unittest.TestCase):
    # Put generated files and folders in this directory.
    OUTPUT_DIRECTORY = "tests/rllib_train_test/"

    @classmethod
    def setUpClass(cls):
        log_dir = os.path.join(RLlibTrainTest.OUTPUT_DIRECTORY, "tests/rllib_results/")
        if os.path.exists(log_dir):
            shutil.rmtree(log_dir)

    def test_rllib_train_cli(self):
        ray.shutdown()
        log_dir = os.path.join(RLlibTrainTest.OUTPUT_DIRECTORY, "tests/rllib_results/")
        try:
            os.system(
                f"python ultra/rllib_train.py --task 00 --level easy --episodes 1 --training-batch-size 200 --headless True --log-dir {log_dir}"
            )
        except Exception as err:
            print(err)
            self.assertTrue(False)

        if os.path.exists(log_dir):
            self.assertTrue(True)
        else:
            self.assertTrue(False)

    def test_rllib_train_method(self):
        log_dir = os.path.join(RLlibTrainTest.OUTPUT_DIRECTORY, "tests/rllib_results/")
        try:
            ray.shutdown()
            ray.init()
            train(
                task=("00", "easy"),
                num_episodes=1,
                policy="ppo",
                eval_info={
                    "eval_rate": 2,
                    "eval_episodes": 1,
                },
                timestep_sec=0.1,
                headless=True,
                seed=2,
                training_batch_size=200,
                log_dir=log_dir,
            )
        except Exception as err:
            print(err)
            self.assertTrue(False)

        if os.path.exists(log_dir):
            self.assertTrue(True)
        else:
            self.assertTrue(False)

    @classmethod
    def tearDownClass(cls):
        if os.path.exists(RLlibTrainTest.OUTPUT_DIRECTORY):
            shutil.rmtree(RLlibTrainTest.OUTPUT_DIRECTORY)



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import argparse
import copy
import glob
import json
import os
import random
import re
import timeit
from ast import literal_eval
from collections import defaultdict

import dill
import gym
import numpy as np
import ray
import torch
from matplotlib import pyplot as plt

from smarts.core.agent import Agent, AgentSpec
from smarts.core.agent_interface import AgentInterface, AgentType, NeighborhoodVehicles
from smarts.core.controllers import ActionSpaceType
from ultra.scenarios.common.social_vehicle_definitions import get_social_vehicle_color
from ultra.scenarios.common.visualization import (
    convert_to_gif,
    draw_intersection,
    profile_vehicles,
)


class DefaultPolicy(Agent):
    def act(self, obs):
        lane_index = 0
        num_trajectory_points = min([10, len(obs.waypoint_paths[lane_index])])
        desired_speed = 20
        trajectory = [
            [
                obs.waypoint_paths[lane_index][i].pos[0]
                for i in range(num_trajectory_points)
            ],
            [
                obs.waypoint_paths[lane_index][i].pos[1]
                for i in range(num_trajectory_points)
            ],
            [
                obs.waypoint_paths[lane_index][i].heading
                for i in range(num_trajectory_points)
            ],
            [desired_speed for i in range(num_trajectory_points)],
        ]
        return trajectory


class BaseAnalysis:
    def __init__(self):
        self.social_vehicles_states = {}
        self.social_vehicles_ids = set()
        self.finished_social_vehicles_ids = set()
        self.analysis = {}

    def reset_scenario_cache(self):
        self.social_vehicles_ids = {}
        self.finished_social_vehicles_ids = set()
        self.social_vehicles_states = defaultdict(
            lambda: {
                "position": [0, 0],
                "route": (),
                "in_junction": [],
                "behavior": None,
                "steps": 0,
                "stop_step": 0,
                "start_step": None,
                "end_step": None,
                "edge": None,
                "speeds": [0],
                "accels": [0],
            }
        )

    def save_histogram(self, data, figure_name, title, x_range=None):
        plt.figure()

        n, bins, patches = plt.hist(
            x=data, bins=600, color="#0504aa", alpha=0.7, rwidth=0.85
        )
        plt.grid(axis="y", alpha=0.75)
        plt.xlabel(title)
        plt.ylabel("Frequency")
        maxfreq = n.max()
        # Set a clean upper y-axis limit.
        plt.ylim(ymax=np.ceil(maxfreq / 10) * 10 if maxfreq % 10 else maxfreq + 10)
        if x_range is not None:
            plt.xlim(x_range)

        plt.savefig(figure_name)
        plt.close()

    def get_agent(self, ego, policy, max_episode_steps):
        observation_adapter = None
        if ego:
            config = get_agent_config_by_type(policy)
            agent_spec = AgentSpec(
                interface=config["interface"],
                policy_params=dict(config["policy"], checkpoint_dir=ego_model),
                policy_builder=config["policy_class"],
            )
            agent_spec.interface.max_episode_steps = max_episode_steps
            observation_adapter = IntersectionAdapter(
                agent_id="AGENT_007",
                social_vehicle_config=config["social_vehicle_config"],
                timestep_sec=config["env"]["timestep_sec"],
                **config["other"],
            )
        else:
            # Lane Following agent
            agent_spec = AgentSpec(
                interface=AgentInterface(
                    max_episode_steps=max_episode_steps,  # 10 mins
                    waypoints=True,
                    action=ActionSpaceType.Lane,
                    debug=False,
                    neighborhood_vehicles=NeighborhoodVehicles(radius=2000),
                ),
                agent_builder=DefaultPolicy,
            )
        return agent_spec, observation_adapter

    def process_social_vehicles(self, vehicles, dt, step):
        stopwatcher_state = None
        stopwatcher_exit = False
        current_vehicles = set()
        for v in vehicles:
            v_id = v.id
            is_new = False
            behavior_key, _ = get_social_vehicle_color(v.id)

            if v_id not in self.social_vehicles_ids:
                self.social_vehicles_ids[v_id] = len(self.social_vehicles_ids)
                is_new = True
            current_vehicles.add(self.social_vehicles_ids[v_id])
            # -----------------------------------
            # 1- Find route information
            # -----------------------------------
            _route = tuple(re.findall("edge-((....|.....)-..)", v_id))
            route = (_route[0][0], _route[1][0])

            previous_edge = self.social_vehicles_states[self.social_vehicles_ids[v_id]][
                "edge"
            ]
            current_edge = re.findall("edge-(....|.....|junction*)-", v.edge_id)
            current_edge = current_edge[0] if len(current_edge) == 1 else None

            # -----------------------------------
            # 2- start caching
            # -----------------------------------
            if is_new:
                self.social_vehicles_states[self.social_vehicles_ids[v_id]][
                    "start_step"
                ] = step
                self.social_vehicles_states[self.social_vehicles_ids[v_id]][
                    "behavior"
                ] = behavior_key
                self.social_vehicles_states[self.social_vehicles_ids[v_id]][
                    "route"
                ] = route

            # -----------------------------------
            # 3- apply updates
            # -----------------------------------
            # update position
            self.social_vehicles_states[self.social_vehicles_ids[v_id]]["position"] = [
                v.position[0],
                v.position[1],
            ]

            # stopped time
            if v.speed < 0.01:
                self.social_vehicles_states[self.social_vehicles_ids[v_id]][
                    "stop_step"
                ] += 1

            # collect speed and accel:
            accel = (
                v.speed
                - self.social_vehicles_states[self.social_vehicles_ids[v_id]]["speeds"][
                    -1
                ]
            ) / dt
            self.social_vehicles_states[self.social_vehicles_ids[v_id]][
                "speeds"
            ].append(v.speed)
            self.social_vehicles_states[self.social_vehicles_ids[v_id]][
                "accels"
            ].append(accel)

            # in junction time range
            if "junction" in v.edge_id:
                n = len(
                    self.social_vehicles_states[self.social_vehicles_ids[v_id]][
                        "speeds"
                    ]
                )
                if not self.social_vehicles_states[self.social_vehicles_ids[v_id]][
                    "in_junction"
                ]:
                    self.social_vehicles_states[self.social_vehicles_ids[v_id]][
                        "in_junction"
                    ] = [n, n]
                else:
                    self.social_vehicles_states[self.social_vehicles_ids[v_id]][
                        "in_junction"
                    ][0] = min(
                        self.social_vehicles_states[self.social_vehicles_ids[v_id]][
                            "in_junction"
                        ][0],
                        n,
                    )
                    self.social_vehicles_states[self.social_vehicles_ids[v_id]][
                        "in_junction"
                    ][1] = max(
                        self.social_vehicles_states[self.social_vehicles_ids[v_id]][
                            "in_junction"
                        ][1],
                        n,
                    )

            self.social_vehicles_states[self.social_vehicles_ids[v_id]]["steps"] += 1
            if current_edge:
                self.social_vehicles_states[self.social_vehicles_ids[v_id]][
                    "edge"
                ] = current_edge

            # -------------------------------------
            # 3- end_time
            # -----------------------------------
            self.social_vehicles_states[self.social_vehicles_ids[v_id]][
                "end_step"
            ] = step

            # check for stopwatcher vehicle
            if (
                "stopwatcher"
                in self.social_vehicles_states[self.social_vehicles_ids[v_id]][
                    "behavior"
                ]
            ):
                stopwatcher_state = self.social_vehicles_states[
                    self.social_vehicles_ids[v_id]
                ]

        # cleanup expired vehicles
        for v in list(self.social_vehicles_states.keys()):
            if v not in current_vehicles:
                self.finished_social_vehicles_ids.add(v)
                if "stopwatcher" in self.social_vehicles_states[v]["behavior"]:
                    stopwatcher_exit = True

        return stopwatcher_state, stopwatcher_exit

    # Number of GPUs should be splited between remote functions.
    def run(
        self,
        scenarios,
        timestep_sec,
        ego,
        policy,
        max_episode_steps,
        video_rate,
        custom_traci_functions,
        end_by_stopwatcher,
        analyze_func,
        init_time_skip=20,
    ):
        agent_spec, observation_adapter = self.get_agent(ego, policy, max_episode_steps)
        agent_id = "AGENT_007"
        ego_less = False if ego else True
        env = gym.make(
            "smarts.env:hiway-v0",
            scenarios=scenarios,
            agent_specs={agent_id: agent_spec},
            headless=True,
            visdom=False,
            timestep_sec=timestep_sec,
            sumo_headless=True,
            seed=0,
            endless_traffic=False,
        )
        visited_scenario = set()
        while len(visited_scenario) < len(scenarios):
            agent = agent_spec.build_agent()
            observations = env.reset()
            scenario_path = env._smarts.scenario.root_filepath

            # only run scenario once
            if scenario_path not in visited_scenario:
                with open(f"{scenario_path}/metadata.json", "r") as metadata_rd:
                    metadata = json.load(metadata_rd)
                visited_scenario.add(scenario_path)
                all_waypoints = [
                    [linked_wp.wp.pos[0], linked_wp.wp.pos[1]]
                    for linked_wp in env._smarts.waypoints._linked_waypoints
                ]
                dones = {"__all__": False}
                intersection_name, intersection_tag = (
                    metadata["intersection"]["name"],
                    metadata["intersection"]["type"],
                )
                total_vehicles = int(metadata["total_vehicles"])
                print(
                    f"Processing Scenario {len(visited_scenario)}/{len(scenarios)}: {scenario_path} "
                )
                print(f"Total number of vehicles: {total_vehicles}")
                print(f"Running egoless: {ego_less}")

                self.reset_scenario_cache()
                step, episode_time = 0, 0.0
                simulation_start = timeit.default_timer()
                stopwatcher_exit = False
                stopwatcher_logged = False
                stopwatcher_max_steps = 0
                images, simulation_data = [], {}
                while not dones["__all__"]:
                    if step % 200 == 0:
                        print("step", step)
                    # todo ------ add ego agent
                    agent_obs = observations[agent_id]
                    start = agent_obs.ego_vehicle_state.mission.start
                    goal = agent_obs.ego_vehicle_state.mission.goal
                    path = agent_obs.waypoint_paths[0]
                    agent_action = "slow_down"
                    observations, rewards, dones, infos = env.step(
                        {agent_id: agent_action}
                    )
                    # ------------
                    (
                        stopwatcher_state,
                        stopwatcher_exit,
                    ) = self.process_social_vehicles(
                        agent_obs.neighborhood_vehicle_states,
                        timestep_sec,
                        step,
                    )
                    if stopwatcher_state:
                        has_stopwatcher = True
                        stopwatcher_max_steps = max(
                            stopwatcher_max_steps, stopwatcher_state["steps"]
                        )
                        if not stopwatcher_logged:
                            print("stopwatcher detected!")
                            stopwatcher_logged = True
                    if (
                        step > (init_time_skip / timestep_sec)
                        and step % int(1 / timestep_sec) == 0
                    ):
                        for func in custom_traci_functions:
                            func(
                                env._smarts.traffic_sim._traci_conn,
                                simulation_data,
                                last_step=step,
                            )

                    step += 1
                    episode_time += timestep_sec

                    if len(visited_scenario) % video_rate == 0:
                        images.append(
                            draw_intersection(
                                ego_position=agent_obs.ego_vehicle_state.position,
                                social_vehicle_states=agent_obs.neighborhood_vehicle_states,
                                goal_path=path,
                                all_waypoints=all_waypoints,
                                step=step,
                                lookaheads_positions=[
                                    waypoint.pos for waypoint in path
                                ],
                                goal=goal.position[0:2],
                                start=start.position[0:2],
                                intersection_tag=intersection_tag,
                                finished_vehicles=self.finished_social_vehicles_ids,
                            )
                        )
                    if ego_less and stopwatcher_exit and end_by_stopwatcher:
                        break
                    if len(self.finished_social_vehicles_ids) == total_vehicles:
                        break
                simulation_end = timeit.default_timer()
                if ego_less and stopwatcher_max_steps > 0:
                    print("Saved stopwatcher steps to", scenario_path)
                    with open(f"{scenario_path}/max_steps.txt", "w") as file_wr:
                        file_wr.write(str(stopwatcher_max_steps))
                simulation_time = simulation_end - simulation_start
                print(analyze_func)
                analyze_func(episode_time, stopwatcher_max_steps, simulation_data)
                print(
                    f"Episode finished in {simulation_time}s, {episode_time}s simulation"
                )

                if len(visited_scenario) % video_rate == 0:
                    print("Converting images to video...")
                    convert_to_gif(
                        images=images,
                        save_dir=scenario_path,
                        name=f"{intersection_name}-{metadata['flow_id']}",
                    )

                print("--------------------------------")

        env.close()



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import json
import os
import shutil
import sys
import unittest

import gym
import ray

from smarts.core.agent import AgentSpec
from smarts.zoo.registry import make
from ultra.baselines.sac.sac.policy import SACPolicy
from ultra.train import train

AGENT_ID = "001"
seed = 2


class TrainTest(unittest.TestCase):
    # Put generated files and folders in this directory.
    OUTPUT_DIRECTORY = "tests/train_test/"

    def test_train_cli(self):
        log_dir = os.path.join(TrainTest.OUTPUT_DIRECTORY, "logs/")
        try:
            os.system(
                f"python ultra/train.py --task 00 --level easy --episodes 1 --max-episode-steps 2 --log-dir {log_dir}"
            )
        except Exception as err:
            print(err)
            self.assertTrue(False)

        if os.path.exists(log_dir):
            self.assertTrue(True)

    def test_train_cli_multiagent(self):
        log_dir = os.path.join(TrainTest.OUTPUT_DIRECTORY, "logs/")
        try:
            os.system(
                f"python ultra/train.py --task 00-multiagent --level easy --episodes 1 --max-episodes-steps 2 --log-dir {log_dir} --policy dqn,bdqn,ppo"
            )
        except Exception as err:
            print(err)
            self.assertTrue(False)

        if os.path.exists(log_dir):
            self.assertTrue(True)

    def test_train_single_agent(self):
        log_dir = os.path.join(TrainTest.OUTPUT_DIRECTORY, "logs/")
        if os.path.exists(log_dir):
            shutil.rmtree(log_dir)

        seed = 2
        policy_classes = ["ultra.baselines.sac:sac-v0"]

        ray.shutdown()
        try:
            ray.init(ignore_reinit_error=True)
            ray.wait(
                [
                    train.remote(
                        scenario_info=("00", "easy"),
                        policy_classes=policy_classes,
                        num_episodes=1,
                        max_episode_steps=2,
                        eval_info={
                            "eval_rate": 1000,
                            "eval_episodes": 2,
                        },
                        timestep_sec=0.1,
                        headless=True,
                        seed=2,
                        log_dir=log_dir,
                    )
                ]
            )
            ray.shutdown()
            self.assertTrue(True)
        except ray.exceptions.WorkerCrashedError as err:
            print(err)
            self.assertTrue(False)
            ray.shutdown()

    def test_train_multiagent(self):
        log_dir = os.path.join(TrainTest.OUTPUT_DIRECTORY, "logs/")
        if os.path.exists(log_dir):
            shutil.rmtree(log_dir)

        seed = 2
        policy_classes = [
            "ultra.baslines.sac:sac-v0",
            "ultra.baselines.ppo:ppo-v0",
            "ultra.baselines.td3:td3-v0",
        ]

        ray.shutdown()
        try:
            ray.init(ignore_reinit_error=True)
            ray.wait(
                [
                    train.remote(
                        scenario_info=("00-multiagent", "easy"),
                        policy_classes=policy_classes,
                        num_episodes=1,
                        max_episode_steps=2,
                        eval_info={
                            "eval_rate": 1000,
                            "eval_episodes": 2,
                        },
                        timestep_sec=0.1,
                        headless=True,
                        seed=2,
                        log_dir=log_dir,
                    )
                ]
            )
            ray.shutdown()
            self.assertTrue(True)
        except ray.exceptions.WorkerCrashedError as err:
            print(err)
            self.assertTrue(False)
            ray.shutdown()

    def test_check_agents_from_pool(self):
        seed = 2
        policy = ""

        with open("ultra/agent_pool.json") as f:
            data = json.load(f)
            for policy in data["agents"].keys():
                policy_path = data["agents"][policy]["path"]
                policy_locator = data["agents"][policy]["locator"]
                policy_class = str(policy_path) + ":" + str(policy_locator)
                try:
                    spec = make(locator=policy_class)
                    agent = spec.build_agent()
                except ImportError as err:
                    self.assertTrue(False)

    def test_spec_is_instance_agentspec(self):
        policy_class = "ultra.baselines.sac:sac-v0"
        spec = make(locator=policy_class)
        self.assertIsInstance(spec, AgentSpec)

    def test_agent_is_instance_policy(self):
        policy_class = "ultra.baselines.sac:sac-v0"
        spec = make(locator=policy_class)
        agent = spec.build_agent()
        self.assertIsInstance(agent, SACPolicy)

    def tearDown(self):
        log_dir = os.path.join(TrainTest.OUTPUT_DIRECTORY, "logs/")
        if os.path.exists(log_dir):
            shutil.rmtree(log_dir)

    @classmethod
    def tearDownClass(cls):
        if os.path.exists(TrainTest.OUTPUT_DIRECTORY):
            shutil.rmtree(TrainTest.OUTPUT_DIRECTORY)



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# code integrated from
#  1- https://github.com/udacity/deep-reinforcement-learning
#  2- https://github.com/sfujim/TD3/blob/master/TD3.py
#

import numpy as np
import torch
import torch.nn.functional as F


def hidden_init(layer):
    fan_in = layer.weight.data.size()[0]
    lim = 1.0 / np.sqrt(fan_in)
    return (-lim, lim)


class ActorNetwork(torch.nn.Module):
    def __init__(
        self,
        input_channels,
        low_dim_state_size,
        action_space,
        seed,
        hidden_dim=512,
    ):
        super(ActorNetwork, self).__init__()
        self.state_space = low_dim_state_size
        self.action_space = action_space
        self.seed = torch.manual_seed(seed)

        self.conv1 = torch.nn.Conv2d(
            input_channels, 32, 8, stride=4
        )  # 84-8+4/4 -> 20x20x32
        self.conv2 = torch.nn.Conv2d(32, 64, 4, stride=2)  # 24-3+2/2 -> 9x9x64
        self.conv3 = torch.nn.Conv2d(64, 64, 3, stride=1)  # 9-3+1/1 -> 7x7x64
        self.fc1 = torch.nn.Linear(7 * 7 * 64 + low_dim_state_size, hidden_dim)
        self.fc2 = torch.nn.Linear(hidden_dim, action_space)
        self.reset_parameters()

    def reset_parameters(self):
        self.conv1.weight.data.uniform_(*hidden_init(self.conv1))
        self.conv2.weight.data.uniform_(*hidden_init(self.conv2))
        self.conv3.weight.data.uniform_(*hidden_init(self.conv3))
        self.fc1.weight.data.uniform_(*hidden_init(self.fc1))
        self.fc2.weight.data.uniform_(-1e-3, 1e-3)

    def forward(self, image, low_dim_state):
        output = F.relu(self.conv1(image))
        output = F.relu(self.conv2(output))
        output = F.relu(self.conv3(output))
        output = output.view(output.size(0), -1)
        output = torch.cat((output, low_dim_state), dim=1)
        output = F.relu(self.fc1(output))
        return self.fc2(output)  # remove torch.tanh


class CriticNetwork(torch.nn.Module):
    def __init__(
        self,
        input_channels,
        low_dim_state_size,
        action_space,
        seed,
        hidden_dim=512,
    ):
        super(CriticNetwork, self).__init__()
        self.seed = torch.manual_seed(seed)
        self.state_space = low_dim_state_size
        self.action_space = action_space

        self.conv1 = torch.nn.Conv2d(
            input_channels, 32, 8, stride=4
        )  # 100-8+4/4 -> 24x24x32
        self.conv2 = torch.nn.Conv2d(32, 64, 4, stride=2)  # 24-3+2/2 -> 11x11x64
        self.conv3 = torch.nn.Conv2d(64, 64, 3, stride=1)  # 9-3+1/1 -> 9x9x64
        self.fc1 = torch.nn.Linear(
            7 * 7 * 64 + low_dim_state_size + action_space, hidden_dim
        )
        self.fc2 = torch.nn.Linear(hidden_dim, 1)

    def reset_parameters(self):
        self.conv1.weight.data.uniform_(*hidden_init(self.conv1))
        self.conv2.weight.data.uniform_(*hidden_init(self.conv2))
        self.conv3.weight.data.uniform_(*hidden_init(self.conv3))
        self.fc1.weight.data.uniform_(*hidden_init(self.fc1))
        self.fc2.weight.data.uniform_(-1e-3, 1e-3)

    def forward(self, image, low_dim_state, action):
        output = F.relu(self.conv1(image))
        output = F.relu(self.conv2(output))
        output = F.relu(self.conv3(output))
        output = output.view(output.size(0), -1)
        output = torch.cat((output, low_dim_state, action), dim=1)
        output = F.relu(self.fc1(output))
        return self.fc2(output)



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import ray.rllib.agents.ppo as ppo
import ray.rllib.agents.sac as sac
import ray.rllib.agents.ddpg as ddpg


class RllibAgent:
    def __init__(self, agent_name, env, config, logger_creator):
        assert agent_name in [
            "td3",
            "ddpg",
            "ppo",
        ], "Some policies are not currently supported (dqn,sac)"  # dqn and sac not currently supported
        self.agent_name = agent_name

        if self.agent_name == "ppo":
            self.trainer = ppo.PPOTrainer(
                env=env,
                config=config,
                logger_creator=logger_creator,
            )
        elif self.agent_name == "ddpg":
            self.trainer = ddpg.DDPGTrainer(
                env=env,
                config=config,
                logger_creator=logger_creator,
            )
        elif self.agent_name == "td3":
            self.trainer = ddpg.TD3Trainer(
                env=env,
                config=config,
                logger_creator=logger_creator,
            )
        # elif self.agent_name=='dqn':
        #     trainer = dqn.DQNTrainer(
        #         env=RLlibUltraEnv,
        #         config=tune_config,
        #         logger_creator=log_creator(log_dir),
        #     )

    def train(self):
        return self.trainer.train()

    def log_evaluation_metrics(self, results):
        return self.trainer.log_result(results)

    @staticmethod
    def rllib_default_config(agent_name):
        assert agent_name in [
            "td3",
            "ddpg",
            "ppo",
        ], "Some policies are not currently supported (dqn,sac)"  # dqn and sac not currently supported

        if agent_name == "ppo":
            return ppo.DEFAULT_CONFIG.copy()
        elif agent_name == "ddpg":
            return ddpg.DEFAULT_CONFIG.copy()
        elif agent_name == "td3":
            return td3.DEFAULT_CONFIG.copy()



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import torch
import torch.nn as nn
import numpy as np


class PrecogFeatureExtractor(nn.Module):
    def __init__(
        self,
        hidden_units=64,
        n_social_features=None,
        social_capacity=None,
        embed_dim=None,
        seed=None,
    ):
        super(PrecogFeatureExtractor, self).__init__()
        if seed is not None:
            self.seed = torch.manual_seed(seed)

        self.social_capacity = social_capacity
        self.embed_dim = embed_dim
        self.n_social_features = n_social_features
        self.social_net = nn.Sequential(
            nn.Linear(n_social_features, hidden_units),
            nn.ReLU(),
            nn.Linear(hidden_units, hidden_units),
            nn.ReLU(),
            nn.Linear(hidden_units, embed_dim),
        )

        self.output_dim = embed_dim * social_capacity

    def forward(self, social_states, training=False):
        if self.social_capacity == 0:
            social_states = [torch.empty(0, self.n_social_features)]
        else:
            social_states = [e[: self.social_capacity] for e in social_states]

        social_lens = [len(e) for e in social_states]
        slice_indices = list(np.cumsum(social_lens))
        slice_indices.insert(0, 0)
        social_features = torch.cat(social_states, 0)
        social_embeddings = self.social_net(social_features)
        masked_embeddings = [
            self._mask_embedding(
                social_embeddings[slice_indices[j] : slice_indices[j + 1]]
            )
            for j in range(len(social_states))
        ]

        state = [e.unsqueeze(0) for e in masked_embeddings]

        return state, {}

    def _mask_embedding(self, embedding):
        device = self.social_net[0].weight.device
        output = torch.zeros(self.social_capacity, self.embed_dim, device=device)
        limit = min(embedding.shape[0], self.social_capacity)
        output[:limit] = embedding[:limit]
        return output.flatten()



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import yaml
import json
import re

loader = yaml.SafeLoader
loader.add_implicit_resolver(
    u"tag:yaml.org,2002:float",
    re.compile(
        u"""^(?:
     [-+]?(?:[0-9][0-9_]*)\\.[0-9_]*(?:[eE][-+]?[0-9]+)?
    |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
    |\\.[0-9_]+(?:[eE][-+][0-9]+)?
    |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\\.[0-9_]*
    |[-+]?\\.(?:inf|Inf|INF)
    |\\.(?:nan|NaN|NAN))$""",
        re.X,
    ),
    list(u"-+0123456789."),
)

# to preserve exponential numbers
def load_yaml(path):
    with open(path, "r") as file:
        data = yaml.safe_load(file)
        json_data = json.dumps(data)
    return yaml.load(json_data, Loader=loader)



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
from ultra.baselines.common.social_vehicle_extraction import *
from ultra.baselines.common.social_vehicles_encoders.pointnet_encoder import PNEncoder
from ultra.baselines.common.social_vehicles_encoders.pointnet_encoder_batched import (
    PNEncoderBatched,
)
from ultra.baselines.common.social_vehicle_extraction import *
from ultra.baselines.common.social_vehicles_encoders.precog_encoder import (
    PrecogFeatureExtractor,
)

social_vehicle_extractors = {
    "pointnet_encoder": extract_social_vehicle_state_pointnet,
    "pointnet_encoder_batched": extract_social_vehicle_state_pointnet,
    "precog_encoder": extract_social_vehicle_state_default,
    "no_encoder": extract_social_vehicle_state_default,
}


def get_social_vehicle_configs(
    encoder_key,
    num_social_features,
    social_capacity,
    seed,
    social_policy_hidden_units=0,
    social_policy_init_std=0,
    **other
):
    config = {
        "num_social_features": int(num_social_features),
        "social_capacity": int(
            social_capacity
        ),  # number of social vehicles to consider
        "social_vehicle_extractor_func": social_vehicle_extractors[encoder_key],
        "encoder_key": encoder_key,
    }
    if encoder_key == "precog_encoder":
        config["encoder"] = {  # state_size: 18 + social_capacity*output_dim
            "use_leading_vehicles": None,
            "social_feature_encoder_class": PrecogFeatureExtractor,
            "social_feature_encoder_params": {
                "hidden_units": int(social_policy_hidden_units),
                "n_social_features": int(num_social_features),
                "embed_dim": 8,
                "social_capacity": int(social_capacity),
                "seed": int(seed),
            },
        }
    elif encoder_key == "pointnet_encoder":
        config["encoder"] = {
            "use_leading_vehicles": None,
            "social_feature_encoder_class": PNEncoder,
            "social_feature_encoder_params": {
                "input_dim": int(num_social_features),
                "nc": 8,
                "transform_loss_weight": 0.1,
            },
        }
    elif encoder_key == "pointnet_encoder_batched":
        config["encoder"] = {
            "use_leading_vehicles": None,
            "social_feature_encoder_class": PNEncoderBatched,
            "social_feature_encoder_params": {
                "input_dim": int(num_social_features),
                "nc": 8,
                "transform_loss_weight": 0.1,
            },
        }
    elif encoder_key == "no_encoder":
        config["encoder"] = {  # state_size: 18 + social_capacity*social_features
            "use_leading_vehicles": {
                "social_capacity": int(social_capacity),
                "max_dist_social_vehicle": 100,
                "num_social_vehicle_per_lane": 2,
            },
            "social_feature_encoder_class": None,  # No Encoder; Examples at the bottom of this page
            "social_feature_encoder_params": {},
        }
    # print(">>>>>", config["encoder"], encoder_key)
    return config



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import glob
import os
import pickle
import re
import shutil
import sys
import unittest

import dill
import gym
import ray

from smarts.core.controllers import ActionSpaceType
from ultra.baselines.agent_spec import BaselineAgentSpec
from ultra.baselines.sac.sac.policy import SACPolicy
from ultra.evaluate import evaluate, evaluation_check
from ultra.utils.episode import episodes

seed = 2
AGENT_ID = "001"


class EvaluateTest(unittest.TestCase):
    # Put generated files and folders in this directory.
    OUTPUT_DIRECTORY = "tests/evaluate_test/"

    @classmethod
    def setUpClass(cls):
        path = os.path.join(EvaluateTest.OUTPUT_DIRECTORY, "sac_test_models/")
        multiagent_path = os.path.join(
            EvaluateTest.OUTPUT_DIRECTORY, "multiagent_test_models/"
        )
        generate_command = (
            "python ultra/scenarios/interface.py generate "
            "--task 00 --level eval_test --root-dir tests/scenarios "
            " --save-dir tests/task/eval_test/eval"
        )
        multiagent_generate_command = (
            "python ultra/scenarios/interface.py generate "
            "--task 00-multiagent --level eval_test --root-dir tests/scenarios "
            "--save-dir tests/task/eval_test_multiagent/eval"
        )
        train_command = (
            "python ultra/train.py "
            "--task 00 --level eval_test --policy sac --headless True --episodes 1 "
            f"--eval-rate 1 --eval-episodes 1 --max-episode-steps 2 --log-dir {path}"
        )
        multiagent_train_command = (
            "python ultra/train.py "
            "--task 00-multiagent --level eval_test --policy sac,dqn,ppo --headless True --episodes 1 "
            f"--eval-rate 1 --eval-episodes 1 --max-episode-steps 2 --log-dir {multiagent_path}"
        )

        # Generate the scenarios.
        os.system(generate_command)
        os.system(multiagent_generate_command)

        # Remove existing models
        if os.path.exists(path):
            shutil.rmtree(path)
        if os.path.exists(multiagent_path):
            shutil.rmtree(multiagent_path)

        # Generate models before evaluation tests
        if not os.path.exists(path):
            os.system(train_command)
        if not os.path.exists(multiagent_path):
            os.system(multiagent_train_command)

    def test_a_folders(self):
        path = os.path.join(EvaluateTest.OUTPUT_DIRECTORY, "sac_test_models/")
        if not os.path.exists(path):
            self.assertTrue(False)

        path = glob.glob(
            os.path.join(EvaluateTest.OUTPUT_DIRECTORY, "sac_test_models/*/models")
        )[0]
        if len(os.listdir(path)) == 0:
            self.assertTrue(False)

        path = "tests/task/eval_test"
        if len(os.listdir(path)) <= 2:
            self.assertTrue(False)

        multiagent_path = os.path.join(
            EvaluateTest.OUTPUT_DIRECTORY, "multiagent_test_models/"
        )
        if not os.path.exists(path):
            self.assertTrue(False)

        multiagent_path = glob.glob(
            os.path.join(
                EvaluateTest.OUTPUT_DIRECTORY, "multiagent_test_models/*/models"
            )
        )[0]
        if len(os.listdir(multiagent_path)) < 2:
            self.assertTrue(False)

        multiagent_path = "tests/task/eval_test_multiagent"
        if len(os.listdir(path)) <= 2:
            self.assertTrue(False)

    def test_evaluation_check(self):
        log_dir = os.path.join(EvaluateTest.OUTPUT_DIRECTORY, "output_eval_check_logs/")
        # ray.init(ignore_reinit_error=True)
        ray.shutdown()
        ray.init()
        try:
            run_experiment(
                scenario_info=("00", "eval_test"), num_agents=1, log_dir=log_dir
            )
            self.assertTrue(True)
        except Exception as err:
            print(err)
            self.assertTrue(False)

        if not os.listdir(log_dir):
            raise "Evaluation failed to generate new experiment folder"
            self.assertTrue(False)
        else:
            shutil.rmtree(log_dir)

    def test_evaluation_check_multiagent(self):
        log_dir = os.path.join(
            EvaluateTest.OUTPUT_DIRECTORY, "output_eval_check_multiagent_logs/"
        )
        # ray.init(ignore_reinit_error=True)
        ray.shutdown()
        ray.init()
        try:
            run_experiment(
                scenario_info=("00-multiagent", "eval_test"),
                num_agents=3,
                log_dir=log_dir,
            )
            self.assertTrue(True)
        except Exception as err:
            print(err)
            self.assertTrue(False)

        if not os.listdir(log_dir):
            raise "Evaluation failed to generate new experiment folder"
            self.assertTrue(False)
        else:
            shutil.rmtree(log_dir)

    def test_evaluate_cli(self):
        log_dir = os.path.join(EvaluateTest.OUTPUT_DIRECTORY, "output_eval_cli_logs/")
        experiment_dir = glob.glob(
            os.path.join(EvaluateTest.OUTPUT_DIRECTORY, "sac_test_models/*")
        )[0]
        models = " ".join(glob.glob(os.path.join(experiment_dir, "models/*")))
        evaluate_command = (
            f"python ultra/evaluate.py "
            f"--task 00 --level eval_test --models {models} --experiment-dir {experiment_dir} "
            f"--episodes 1 --max-episode-steps 2 --log-dir {log_dir} --headless True"
        )

        ray.shutdown()
        try:
            os.system(evaluate_command)
            self.assertTrue(True)
        except Exception as err:
            print(err)
            self.assertTrue(False)

        if not os.listdir(log_dir):
            raise "Evaluation failed to generate new experiment folder"
            self.assertTrue(False)
        else:
            shutil.rmtree(log_dir)

    def test_evaluate_cli_multiagent(self):
        log_dir = os.path.join(
            EvaluateTest.OUTPUT_DIRECTORY, "output_eval_cli_multiagent_logs/"
        )
        experiment_dir = glob.glob(
            os.path.join(EvaluateTest.OUTPUT_DIRECTORY, "multiagent_test_models/*")
        )[0]
        models = " ".join(glob.glob(os.path.join(experiment_dir, "models/000/")))
        evaluate_command = (
            f"python ultra/evaluate.py "
            f"--task 00-multiagent --level eval_test --models {models} --experiment-dir {experiment_dir} "
            f"--episodes 1 --max-episode-steps 2 --log-dir {log_dir} --headless True"
        )

        ray.shutdown()
        try:
            os.system(evaluate_command)
            self.assertTrue(True)
        except Exception as err:
            print(err)
            self.assertTrue(False)

        if not os.listdir(log_dir):
            raise "Evaluation failed to generate new experiment folder"
            self.assertTrue(False)
        else:
            shutil.rmtree(log_dir)

    def test_evaluate_agent(self):
        seed = 2
        models_directory = glob.glob(
            os.path.join(EvaluateTest.OUTPUT_DIRECTORY, "sac_test_models/*/models/")
        )[0]
        log_dir = os.path.join(EvaluateTest.OUTPUT_DIRECTORY, "output_eval_agent_logs/")

        with open(
            os.path.join(models_directory, "../agent_metadata.pkl"), "rb"
        ) as metadata_file:
            agent_metadata = pickle.load(metadata_file)

        agent_ids = agent_metadata["agent_ids"]
        policy_classes = agent_metadata["agent_classes"]
        checkpoint_directories = {
            agent_id: sorted(
                glob.glob(os.path.join(models_directory, agent_id, "*")),
                key=lambda x: int(x.split("/")[-1]),
            )
            for agent_id in agent_ids
        }

        ray.shutdown()
        ray.init(ignore_reinit_error=True)
        try:
            evaluate.remote(
                experiment_dir=None,
                agent_ids=agent_ids,
                policy_classes=policy_classes,
                seed=seed,
                checkpoint_dirs=checkpoint_directories,
                scenario_info=("00", "eval_test"),
                num_episodes=1,
                max_episode_steps=2,
                timestep_sec=0.1,
                headless=True,
                log_dir=log_dir,
            )
            self.assertTrue(True)
        except Exception as err:
            print(err)
            self.assertTrue(False)

    def test_evaluate_multiagent(self):
        seed = 2
        models_directory = glob.glob(
            os.path.join(
                EvaluateTest.OUTPUT_DIRECTORY, "multiagent_test_models/*/models/"
            )
        )[0]
        log_dir = os.path.join(
            EvaluateTest.OUTPUT_DIRECTORY, "output_eval_multiagent_logs/"
        )

        with open(
            os.path.join(models_directory, "../agent_metadata.pkl"), "rb"
        ) as metadata_file:
            agent_metadata = pickle.load(metadata_file)

        agent_ids = agent_metadata["agent_ids"]
        policy_classes = agent_metadata["agent_classes"]
        checkpoint_directories = {
            agent_id: sorted(
                glob.glob(os.path.join(models_directory, agent_id, "*")),
                key=lambda x: int(x.split("/")[-1]),
            )
            for agent_id in agent_ids
        }

        ray.shutdown()
        ray.init(ignore_reinit_error=True)
        try:
            evaluate.remote(
                experiment_dir=None,
                agent_ids=agent_ids,
                policy_classes=policy_classes,
                seed=seed,
                checkpoint_dirs=checkpoint_directories,
                scenario_info=("00-multiagent", "eval_test"),
                num_episodes=1,
                max_episode_steps=2,
                timestep_sec=0.1,
                headless=True,
                log_dir=log_dir,
            )
            self.assertTrue(True)
        except Exception as err:
            print(err)
            self.assertTrue(False)

    def test_extract_policy_from_path(self):
        paths = [
            "from.ultra.baselines.sac:sac-v0",
            "hello.ultra.ppo:ppo-v1",
            "ultra.custom:custom",
            "a.sb.ultra.c.d.e.sac:sac-v99",
            "a.b.c.d.e.ultra.custom_agent.policy:MBPO-v2",
        ]

        def extract(path):
            m = re.search(
                "ultra(.)*([a-zA-Z0-9_]*.)+([a-zA-Z0-9_])+:[a-zA-Z0-9_]+((-)*[a-zA-Z0-9_]*)*",
                path,
            )

            try:
                policy_class = m.group(0)
            except AttributeError as e:
                self.assertTrue(False)

        for path in paths:
            extract(path)

    # @classmethod
    # def tearDownClass(cls):
    #     os.system("ray stop")

    @classmethod
    def tearDownClass(cls):
        if os.path.exists(EvaluateTest.OUTPUT_DIRECTORY):
            shutil.rmtree(EvaluateTest.OUTPUT_DIRECTORY)
        if os.path.exists("tests/task/eval_test/"):
            shutil.rmtree("tests/task/eval_test/")
        if os.path.exists("tests/task/eval_test_multiagent/"):
            shutil.rmtree("tests/task/eval_test_multiagent/")


def run_experiment(scenario_info, num_agents, log_dir, headless=True):
    agent_ids = ["0" * max(0, 3 - len(str(i))) + str(i) for i in range(num_agents)]
    agent_classes = {agent_id: "ultra.baselines.sac:sac-v0" for agent_id in agent_ids}
    agent_specs = {
        agent_id: BaselineAgentSpec(
            action_type=ActionSpaceType.Continuous,
            policy_class=SACPolicy,
            max_episode_steps=2,
        )
        for agent_id in agent_ids
    }

    env = gym.make(
        "ultra.env:ultra-v0",
        agent_specs=agent_specs,
        scenario_info=scenario_info,
        headless=headless,
        timestep_sec=0.1,
        seed=seed,
    )

    agents = {
        agent_id: agent_spec.build_agent()
        for agent_id, agent_spec in agent_specs.items()
    }

    total_step = 0
    etag = ":".join([policy_class.split(":")[-1] for policy_class in agent_classes])

    for episode in episodes(1, etag=etag, log_dir=log_dir):
        observations = env.reset()
        dones = {"__all__": False}
        infos = None
        episode.reset()
        experiment_dir = episode.experiment_dir

        if not os.path.exists(f"{experiment_dir}/agent_metadata.pkl"):
            if not os.path.exists(experiment_dir):
                os.makedirs(experiment_dir)
            with open(f"{experiment_dir}/agent_metadata.pkl", "wb") as metadata_file:
                dill.dump(
                    {
                        "agent_ids": agent_ids,
                        "agent_classes": agent_classes,
                        "agent_specs": agent_specs,
                    },
                    metadata_file,
                    pickle.HIGHEST_PROTOCOL,
                )

        while not dones["__all__"]:
            evaluation_check(
                agents=agents,
                agent_ids=agent_ids,
                episode=episode,
                eval_rate=10,
                eval_episodes=1,
                max_episode_steps=2,
                policy_classes=agent_classes,
                scenario_info=scenario_info,
                timestep_sec=0.1,
                headless=True,
                log_dir=log_dir,
            )
            actions = {
                agent_id: agents[agent_id].act(observation, explore=True)
                for agent_id, observation in observations.items()
            }
            next_observations, rewards, dones, infos = env.step(actions)

            active_agent_ids = observations.keys() & next_observations.keys()
            loss_outputs = {
                agent_id: agents[agent_id].step(
                    state=observations[agent_id],
                    action=actions[agent_id],
                    reward=rewards[agent_id],
                    next_state=next_observations[agent_id],
                    done=dones[agent_id],
                    info=infos[agent_id],
                )
                for agent_id in active_agent_ids
            }

            episode.record_step(
                agent_ids_to_record=active_agent_ids,
                infos=infos,
                rewards=rewards,
                total_step=total_step,
                loss_outputs=loss_outputs,
            )

            total_step += 1
            observations = next_observations

    env.close()



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""
source code: https://github.com/ShangtongZhang/DeepRL/deep_rl/component/random_process.py
"""
import numpy as np


class LinearSchedule:
    def __init__(self, start, end=None, steps=None):
        if end is None:
            end = start
            steps = 1
        self.inc = (end - start) / float(steps)
        self.current = start
        self.end = end
        if end > start:
            self.bound = min
        else:
            self.bound = max

    def __call__(self, steps=1):
        val = self.current
        self.current = self.bound(self.current + self.inc * steps, self.end)
        return val


class OrnsteinUhlenbeckProcess:
    def __init__(self, size, std, mu, theta=0.15, dt=1e-2, x0=None):
        self.theta = theta
        self.mu = mu
        self.std = std
        self.dt = dt
        self.x0 = x0
        self.size = size
        self.reset_states()

    def sample(self):
        x = (
            self.x_prev
            + self.theta * (self.mu - self.x_prev) * self.dt
            + self.std() * np.sqrt(self.dt) * np.random.randn(*self.size)
        )
        self.x_prev = x
        return x

    def reset_states(self):
        self.x_prev = self.x0 if self.x0 is not None else np.zeros(self.size)



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import os
import re
from collections import defaultdict

import cv2
import imageio
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.backends.backend_agg import FigureCanvasAgg
from matplotlib.figure import Figure

from .social_vehicle_definitions import get_social_vehicle_color

# import geometry as geometry


def convert_to_gif(images, save_dir, name):
    print("Processing images...")
    imageio.mimsave(f"{save_dir}/{name}.gif", images)
    print(f"Saved {save_dir}/{name}.gif")


def convert_to_mov(images, save_dir):
    height, width, layers = images[0].shape
    video_name = f"{save_dir}/movie.mov"
    video = cv2.VideoWriter(video_name, 0, 1, (width, height))
    for image in images:
        bgr = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
        video.write(bgr)


def fig2data(fig_canvas):
    # Retrieve a view on the renderer buffer
    fig_canvas.draw()
    buf = fig_canvas.buffer_rgba()
    # convert to a NumPy array
    image = np.asarray(buf)
    return image


def profile_vehicles(vehicle_states, save_dir):
    def plot(ax, x_label, y_label, state):
        ax.set_xlabel(x_label, fontsize=10)
        ax.set_ylabel(y_label, fontsize=10)
        ax.plot(range(len(state[y_label])), state[y_label])
        if "in_junction" in state:
            ax.axvspan(
                state["in_junction"][0], state["in_junction"][1], color="red", alpha=0.5
            )

    for v_id, state in vehicle_states.items():
        if not state["teleported"]:
            fig, ax = plt.subplots(nrows=1, ncols=2, figsize=(15, 5))
            plot(ax[0], "time", "speed", state)
            plot(ax[1], "time", "accel", state)
            behavior_key, behavior_color = get_social_vehicle_color(state["behavior"])
            temp_save_dir = f"{save_dir}/{behavior_key}"
            if not os.path.exists(temp_save_dir):
                os.makedirs(temp_save_dir)
            plt.savefig(f"{temp_save_dir}/{state['route']}_{v_id}.png")

    plt.close("all")


def draw_intersection(
    ego_position,
    goal_path,
    all_waypoints,
    step,
    goal,
    start,
    social_vehicle_states,
    finished_vehicles,
    lookaheads_positions,
    intersection_tag="t",
):
    fig_w = 812
    fig_h = 812
    canvas = np.zeros((fig_w, fig_h, 3), np.uint8)

    if intersection_tag == "c":
        fig_offset_y = 320
        fig_offset_x = 100
    else:
        fig_offset_y = 700
        fig_offset_x = 300
    fig_offset_y = 100
    fig_offset_x = 100
    fig_mul = 2
    # Create a figure to contain the plot.
    figure = plt.figure(figsize=(20, 10))
    for point in all_waypoints:
        canvas = cv2.circle(
            canvas,
            (
                fig_offset_x + int(point[0] * fig_mul),
                fig_offset_y - int(point[1] * fig_mul),
            ),
            radius=0,
            color=(255, 255, 255),
            thickness=1,
        )
    for point in lookaheads_positions:
        canvas = cv2.circle(
            canvas,
            (
                fig_offset_x + int(point[0] * fig_mul),
                fig_offset_y - int(point[1] * fig_mul),
            ),
            radius=0,
            color=(255, 255, 0),
            thickness=1,
        )
    ego_color = (0, 200, 0)
    colors_legend = set()
    colors_legend.add(("ego", ego_color))
    font = cv2.FONT_HERSHEY_DUPLEX

    for state in social_vehicle_states:  # .items():
        # print(state['behavior'])
        # if v_id not in finished_vehicles:
        # behavior_key, behavior_color = get_social_vehicle_color(state["behavior"])
        pos_x = fig_offset_x + int(state.position[0] * fig_mul)
        pos_y = fig_offset_y - int(state.position[1] * fig_mul)
        canvas = cv2.circle(
            canvas,
            (
                pos_x,
                pos_y,
            ),
            radius=1,
            color=(10, 10, 30),
            thickness=2,
        )
        # canvas = cv2.putText(
        #     canvas, str(v_id), (pos_x + 4, pos_y + 4), font, 0.3, behavior_color, 1,
        # )
        # colors_legend.add((behavior_key, behavior_color))
    # print(ego)
    # if ego:
    canvas = cv2.circle(
        canvas,
        (
            fig_offset_x + int(ego_position[0] * fig_mul),
            fig_offset_y - int(ego_position[1] * fig_mul),
        ),
        radius=1,
        color=ego_color,
        thickness=2,
    )

    canvas = cv2.putText(
        canvas,
        "x",
        (
            fig_offset_x + int(goal[0] * fig_mul) - 4,
            fig_offset_y - int(goal[1] * fig_mul) + 8,
        ),
        font,
        0.5,
        (200, 150, 255),
        1,
    )
    canvas = cv2.putText(
        canvas,
        "x",
        (
            fig_offset_x + int(start[0] * fig_mul) - 4,
            fig_offset_y - int(start[1] * fig_mul) + 8,
        ),
        font,
        0.5,
        (1, 255, 255),
        1,
    )

    # color_offset = 20
    # legend_position = (30, 50)
    #
    # for color_key, color in colors_legend:
    #     canvas = cv2.putText(
    #         canvas,
    #         color_key,
    #         (legend_position[0], legend_position[1] + color_offset),
    #         font,
    #         0.5,
    #         color,
    #         1,
    #     )
    #     color_offset += 20
    cv2.imwrite(f"temp/{step}.png", canvas)
    plt.close("all")
    return canvas


# def visualize_social_safety(ax, all_waypoints, nighbor_bounding_box):
#     all_waypoints = np.asarray(all_waypoints)

#     ax.clear()
#     ax.set_axis_off()
#     ax.view_init(60, 60)
#     ax.scatter(
#         all_waypoints[:, 0],
#         all_waypoints[:, 1],
#         -5.0,
#         s=100.0,
#         marker=".",
#         color="blue",
#         alpha=0.1,
#     )

#     ax = geometry.visualize_boxes(
#         ax,
#         nighbor_bounding_box,
#         ["o" for box in nighbor_bounding_box],
#         [
#             "r" if i == 0 else "g" if i == 1 else "b"
#             for i, box in enumerate(nighbor_bounding_box)
#         ],
#         [
#             "cyan" if i == 0 or i == 1 else "pink"
#             for i, box in enumerate(nighbor_bounding_box)
#         ],
#     )
#     plt.draw()
#     plt.pause(0.00001)



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import matplotlib

matplotlib.use("Agg")
import argparse
import glob
import os
import sys
import time
import timeit

import _pickle
import matplotlib.pyplot as plt

if "SUMO_HOME" in os.environ:
    tools = os.path.join(os.environ["SUMO_HOME"], "tools")
    sys.path.append(tools)
else:
    sys.exit("please declare environment variable 'SUMO_HOME'")
import traci
import traci.constants as tc

v_types = ["default", "aggressive", "slow", "blocker", "crusher"]


def get_vehicle_type(vehicle_name):
    type_id = traci.vehicle.getTypeID(vehicle_name)
    for t in v_types:
        if t in type_id:
            return t


def _base_sumo_load_params(net_file, route_file, delta_t):
    load_params = [
        "--lanechange.duration=3.0",
        "--step-length=%f" % delta_t,
        "--begin=0",
        "--end=31536000",
        #
        "--num-clients=1",
        "--net-file=%s" % net_file,
        "--quit-on-end",
        # "--no-step-log",
        # "--no-warnings=1",
        "--seed=%s" % 354354,
        "--time-to-teleport=%s" % -1,
        "--collision.check-junctions=true",
        "--collision.action=none",
        "--log=%s" % "sumo_log.log"
        #
    ]
    load_params.append("--start")
    load_params.append("--route-files={}".format(route_file))

    return load_params


def get_simulation_info(net_file, route_file, delta_t=1.0):
    params = _base_sumo_load_params(net_file, route_file, delta_t)
    traci.start(["sumo", *params])
    time.sleep(0.1)

    warmup = True
    t = 0.0

    vehicle_stats = {}
    # exited = {}

    while warmup or traci.vehicle.getIDCount() > 0:
        traci.simulationStep(t)
        for id in traci.vehicle.getIDList():
            # if id in exited.keys():
            #     print("Exited reappear")

            if id not in vehicle_stats:
                v_t = get_vehicle_type(id)
                v_r = traci.vehicle.getRoute(id)
                vehicle_stat = {
                    "id": id,
                    "start": t,
                    "end": t,
                    "stop_time": 0,
                    "type": v_t,
                    "route": "{}-{}".format(*v_r),
                }
                vehicle_stats[id] = vehicle_stat
            else:
                vehicle_stats[id]["end"] = t

            if traci.vehicle.getSpeed(id) < 0.01:
                vehicle_stats[id]["stop_time"] += delta_t

            # _exited_id = list(set(vehicle_stats.keys()) - set(traci.vehicle.getIDList()))
            # for _id in _exited_id:
            #     exited[_id] = 1
            # print(len(exited.keys()))

        t += delta_t
        if t > 20:
            warmup = False

    vehicle_types_time_taken = {t: [] for t in v_types}

    routes_time_taken = {
        r: [] for r in list(set([v["route"] for v in vehicle_stats.values()]))
    }
    stop_pcts = {r: [] for r in list(set([v["route"] for v in vehicle_stats.values()]))}

    max_t = 0
    for v in vehicle_stats.values():
        v_t = v["type"]
        v_r = v["route"]
        time_taken = v["end"] - v["start"]
        vehicle_types_time_taken[v_t].append(time_taken)
        routes_time_taken[v_r].append(time_taken)
        _stop_pct = v["stop_time"] / time_taken
        stop_pcts[v_r].append(_stop_pct)
        if v_r != "edge-south-SN-edge-west-EW":
            max_t = max(time_taken, max_t)

    traci.close()

    return {
        "time": t,
        "vehicle_types_time_taken": vehicle_types_time_taken,
        "routes_time_taken": routes_time_taken,
        "max_time": max_t,
        "stop_pcts": stop_pcts,
    }


def save_histogram(time_taken, fn, title):
    plt.figure()
    plt.title(title)
    plt.hist(time_taken, normed=False, bins=20)
    plt.savefig(fn)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("path", type=str, help="wildcard path to scenarios dirs")
    parser.add_argument("--plot_only", "-p", dest="plot_only", action="store_true")
    return parser.parse_args()


if __name__ == "__main__":
    os.makedirs("plots", exist_ok=True)

    args = parse_args()
    plot_only = args.plot_only
    dirs = args.path

    if not plot_only:
        delta_t = 0.1
        failed_scenario = 0

        dirs = sorted(glob.glob(dirs))
        episode_time_taken = []
        vehicle_types_time_taken = {t: [] for t in v_types}
        routes_time_taken = {}
        max_t = []
        stop_pcts = {}

        for j, d in enumerate(dirs):
            start = timeit.default_timer()

            info = get_simulation_info(
                os.path.join(d, "map.net.xml"),
                os.path.join(d, "traffic/all.rou.xml"),
                delta_t,
            )
            _episode_time_taken = info["time"]
            _vehicle_types_time_taken = info["vehicle_types_time_taken"]
            _routes_time_taken = info["routes_time_taken"]
            _max_t = info["max_time"]
            _stop_pcts = info["stop_pcts"]

            episode_time_taken.append(_episode_time_taken)
            for t in v_types:
                vehicle_types_time_taken[t].extend(_vehicle_types_time_taken[t])
            for rn, rt in _routes_time_taken.items():
                if rn not in routes_time_taken:
                    routes_time_taken[rn] = []
                routes_time_taken[rn].extend(rt)
            max_t.append(_max_t)
            for rn, rt in _stop_pcts.items():
                if rn not in stop_pcts:
                    stop_pcts[rn] = []
                stop_pcts[rn].extend(rt)

            print(
                "Scenario {} ({}) take {}s, {}s simulation".format(
                    j, d, timeit.default_timer() - start, _episode_time_taken
                )
            )
    else:
        info = _pickle.load(open(".info.pkl", "rb"))
        episode_time_taken = info["episode_time_taken"]
        vehicle_types_time_taken = info["vehicle_types_time_taken"]
        routes_time_taken = info["routes_time_taken"]
        max_t = info["max_time"]
        stop_pcts = info["stop_pcts"]

    save_histogram(episode_time_taken, "plots/time_episode.png", "time episode")
    for v_t in vehicle_types_time_taken.keys():
        save_histogram(
            vehicle_types_time_taken[v_t],
            "plots/time_type_{}.png".format(v_t),
            "time {}".format(v_t),
        )
    for v_r in routes_time_taken.keys():
        save_histogram(
            routes_time_taken[v_r],
            "plots/time_route_{}.png".format(v_r),
            "time {}".format(v_r),
        )
    save_histogram(
        sum(list(vehicle_types_time_taken.values()), []),
        "plots/time_all.png",
        "time all",
    )
    save_histogram(max_t, "plots/time_max_t.png", "max_t")
    for v_r in stop_pcts.keys():
        save_histogram(
            stop_pcts[v_r],
            "plots/stop_pct_route_{}.png".format(v_r),
            "stop % {}".format(v_r),
        )
    save_histogram(
        sum(list(stop_pcts.values()), []), "plots/stop_pct_all.png", "stop % all"
    )

    _pickle_info = {
        "episode_time_taken": episode_time_taken,
        "vehicle_types_time_taken": vehicle_types_time_taken,
        "routes_time_taken": routes_time_taken,
        "max_time": max_t,
        "stop_pcts": stop_pcts,
    }

    _pickle.dump(_pickle_info, open(".info.pkl", "wb"))



# MIT License
#
# Copyright (C) 2021. Huawei Technologies Co., Ltd. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import torch
from torch import nn


class Flatten(nn.Module):
    def forward(self, x):
        x = x.view(x.size()[0], -1)
        return x


class DQNCNN(nn.Module):
    def __init__(
        self,
        n_in_channels,
        image_dim,
        state_size,
        num_actions,
        hidden_dim=128,
        activation=nn.ReLU,
    ):
        super(DQNCNN, self).__init__()
        self.im_feature = nn.Sequential(
            nn.Conv2d(n_in_channels, 32, 8, 4),
            activation(),
            nn.Conv2d(32, 64, 4, 2),
            activation(),
            nn.Conv2d(64, 64, 3, 1),
            activation(),
            Flatten(),
        )

        dummy = torch.zeros((1, n_in_channels, *image_dim))
        im_feature_size = self.im_feature(dummy).data.cpu().numpy().size

        self.q_outs = nn.ModuleList()
        for num_action in num_actions:
            q_out = nn.Sequential(
                nn.Linear(im_feature_size + state_size, hidden_dim),
                nn.ReLU(),
                nn.Linear(hidden_dim, num_action),
            )
            self.q_outs.append(q_out)

        self.init()

    def init(self):
        for q_out in self.q_outs:
            # nn.init.normal_(q_out[-1].weight.data, 0.0, 1e-2)
            nn.init.constant_(q_out[-1].bias.data, 0.0)

    def forward(self, image, state_size):
        im_feature = self.im_feature(image)
        x = torch.cat([im_feature, state_size], dim=-1)
        x = [e(x) for e in self.q_outs]
        return x


class DQNFC(nn.Module):
    def __init__(
        self,
        num_actions,
        state_size,
        hidden_dim=256,
        activation=nn.ReLU,
    ):
        super(DQNFC, self).__init__()

        self.q_outs = nn.ModuleList()
        for num_action in num_actions:
            q_out = nn.Sequential(
                nn.Linear(state_size, hidden_dim),
                activation(),
                nn.Linear(hidden_dim, hidden_dim // 2),
                activation(),
                nn.Linear(hidden_dim // 2, hidden_dim // 4),
                activation(),
                nn.Linear(hidden_dim // 4, num_action),
            )
            self.q_outs.append(q_out)

        self.init()

    def init(self):
        for q_out in self.q_outs:
            nn.init.constant_(q_out[-1].bias.data, 0.0)
            # nn.init.normal_(q_out[-1].weight.data, 0.0, 1e-3)

    def forward(self, state, training=False):
        low_dim_state = state["low_dim_states"]
        if len(low_dim_state.shape) == 1:
            low_dim_state = torch.unsqueeze(low_dim_state, 0)
            unsqueezed = True
        else:
            unsqueezed = False
        x = low_dim_state
        # x = self.feature(low_dim_state)
        x = [e(x) for e in self.q_outs]
        if unsqueezed:
            x = [torch.squeeze(e, 0) for e in x]

        if training:
            aux_losses = {}
            return x, aux_losses
        else:
            return x


class DQNWithSocialEncoder(nn.Module):
    def __init__(
        self,
        num_actions,
        state_size,
        hidden_dim=256,
        activation=nn.ReLU,
        social_feature_encoder_class=None,
        social_feature_encoder_params=None,
    ):
        super(DQNWithSocialEncoder, self).__init__()

        self.social_feature_encoder = (
            social_feature_encoder_class(**social_feature_encoder_params)
            if social_feature_encoder_class
            else None
        )

        self.state_size = state_size
        self.q_outs = nn.ModuleList()
        for num_action in num_actions:
            q_out = nn.Sequential(
                nn.Linear(state_size, hidden_dim),
                activation(),
                nn.Linear(hidden_dim, hidden_dim // 2),
                activation(),
                nn.Linear(hidden_dim // 2, hidden_dim // 4),
                activation(),
                nn.Linear(hidden_dim // 4, num_action),
            )
            self.q_outs.append(q_out)

        self.init()

    def init(self):
        for q_out in self.q_outs:
            nn.init.constant_(q_out[-1].bias.data, 0.0)

    def forward(self, state, training=False):
        low_dim_state = state["low_dim_states"]
        social_vehicles_state = state["social_vehicles"]

        aux_losses = {}
        social_feature = []
        if self.social_feature_encoder is not None:
            social_feature, social_encoder_aux_losses = self.social_feature_encoder(
                social_vehicles_state, training
            )
            aux_losses.update(social_encoder_aux_losses)
        else:
            social_feature = [e.reshape(1, -1) for e in social_vehicles_state]

        social_feature = torch.cat(social_feature, 0) if len(social_feature) > 0 else []
        x = (
            torch.cat([low_dim_state, social_feature], -1)
            if len(social_feature) > 0
            else low_dim_state
        )
        x = [e(x) for e in self.q_outs]

        if training:
            return x, aux_losses
        else:
            return x
